# coding=utf-8
#
# Copyright (C) 2019 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Common access to serial and other computer ports.
"""

import os
import sys
import time
from .utils import DependencyError, AbortExtension

try:
    import serial
    from serial.tools import list_ports
except ImportError:
    serial = None


class Serial:
    """
    Attempt to get access to the computer's serial port.

    with Serial(port_name, ...) as com:
        com.write(...)

    Provides access to the debug/testing ports which are pretend ports
    able to accept the same input but allow for debugging.
    """

    def __init__(self, port, baud=9600, timeout=0.1, **options):
        self.test = port == "[test]"
        if self.test:
            import pty  # This does not work on windows #pylint: disable=import-outside-toplevel

            self.controller, self.peripheral = pty.openpty()
            port = os.ttyname(self.peripheral)

        self.has_serial()
        self.com = serial.Serial()
        self.com.port = port
        self.com.baudrate = int(baud)
        self.com.timeout = timeout
        self.set_options(**options)

    def set_options(self, stop=1, size=8, flow=None, parity=None):
        """Set further options on the serial port"""
        size = {5: "five", 6: "six", 7: "seven", 8: "eight"}.get(size, size)
        stop = {"onepointfive": 1.5}.get(stop.lower(), stop)
        stop = {1: "one", 1.5: "one_point_five", 2: "two"}.get(stop, stop)
        self.com.bytesize = getattr(serial, str(str(size).upper()) + "BITS")
        self.com.stopbits = getattr(serial, "STOPBITS_" + str(stop).upper())
        self.com.parity = getattr(serial, "PARITY_" + str(parity).upper())
        # set flow control
        self.com.xonxoff = flow == "xonxoff"
        self.com.rtscts = flow in ("rtscts", "dsrdtrrtscts")
        self.com.dsrdtr = flow == "dsrdtrrtscts"

    def __enter__(self):
        try:
            # try to establish connection
            self.com.open()
        except serial.SerialException as error:
            raise AbortExtension(
                "Could not open serial port. Please check your device"
                " is running, connected and the settings are correct"
            ) from error
        return self.com

    def __exit__(self, exc, value, traceback):
        if not traceback and self.test:
            output = " " * 1024
            while len(output) == 1024:
                time.sleep(0.01)
                output = os.read(self.controller, 1024)
                sys.stderr.write(output.decode("utf8"))
        # self.com.read(2)
        self.com.close()

    @staticmethod
    def has_serial():
        """Late importing of pySerial module"""
        if serial is None:
            raise DependencyError("pySerial is required to open serial ports.")

    @staticmethod
    def list_ports():
        """Return a list of available serial ports"""
        Serial.has_serial()  # Cause DependencyError error
        return [hw.name for hw in list_ports.comports(True)]



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2020-2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""XAML Object structure and XAML (LXML) parsing utilities"""


from __future__ import annotations
from typing import Optional, Union, List, Dict, Callable
import random
import re

Number = Union[int, float]

from lxml import etree
import inkex
from inkex.elements._utils import removeNS as inkexRemoveNS

from .export.config import GLOBAL_SETTINGS


class NodeBasedXAMLLookup(etree.PythonElementClassLookup):
    """Lookup for etree custom element classes"""

    def lookup(self, doc, element):  # pylint: disable=unused-argument
        """Lookup called by lxml when assigning elements their object class"""
        try:
            result = globals()[removeNS(element.tag)[-1]]
        except TypeError:
            result = None
        except KeyError:
            result = BaseElement
        return result


XAML_PARSER = etree.XMLParser(huge_tree=True)
XAML_PARSER.set_element_class_lookup(NodeBasedXAMLLookup())


def addNS(tag, ns=None):  # pylint: disable=invalid-name
    """Add a known namespace to a name for use with lxml"""
    return inkex.addNS(tag, ns, GLOBAL_SETTINGS.xaml_namespaces)


def removeNS(name):  # pylint: disable=invalid-name
    """The reverse of addNS, finds any namespace and returns tuple (ns, tag)"""
    return inkexRemoveNS(name, GLOBAL_SETTINGS.xaml_namespaces_reversed, default="xaml")


KEY_ATTR = "x:Key"


class XAMLObject(etree.ElementBase):
    """Base class for XAML objects"""

    PARSER = XAML_PARSER
    NAMESPACE = property(lambda self: GLOBAL_SETTINGS.xaml_namespaces["xaml"])
    TAG = property(lambda self: removeNS(self.classtag())[-1])

    text: str

    @classmethod
    def classtag(cls) -> str:
        """Returns the xml tag of this class"""
        return cls.__name__

    @property
    def tag(self) -> str:
        """Returns the xml tag of this element (i.e. the tag of the elements's class)"""
        return self.__class__.classtag()

    @staticmethod
    def sanitize_name(name):
        """Replace all non-alphanumeric characters; more precise validation would be
        https://docs.microsoft.com/en-us/dotnet/desktop/xaml-services/xamlname-grammar
        also ensure that the first character is not a number"""
        result = re.sub("[^0-9a-zA-Z]", "_", name)
        return re.sub(r"^(\d.*)", r"_\1", result)

    @staticmethod
    def make_key_unique(parent, element: BaseElement, key=KEY_ATTR, force=True):
        """Makes sure that an element does not create a key collision. If so, the key is
        changed to the existing key (or the element tag if not set) + underscore
        + random number"""
        current_key = element.get(key, None)
        if current_key is not None:
            current_key = BaseElement.sanitize_name(current_key)
            element.set(key, current_key)
        newkey = False
        if current_key is None:
            if force:
                newkey = True
                current_key = element.tag
            else:
                # No key is set, there is no danger for collision.
                return

        xpath = f"//*[@{key}='{element.get(key)}']"
        elements = parent.xpath(xpath)
        if len(elements) > 2 or (len(elements) == 1 and elements[0] != element):
            newkey = True
        temp_key = current_key
        while newkey or len(parent.xpath(f"//*[@{key}='{temp_key}']")) > 0:
            temp_key = f"{current_key}_{random.randint(0, 1000)}"
            newkey = False
        element.set(addNS(key), temp_key)

    def get_tag_name(self, prop: str) -> str:
        """Get the tag name for a property"""
        prop = ":".join(removeNS(prop))
        if prop.startswith("xaml:"):
            prop = prop[5:]
        if ":" in prop:
            return prop
        return f"{self.tag}.{prop}"

    def set(self, name, val):
        """Check if the attribute is already set, in this case delete it.
        Otherwise set it."""
        self.remove_property(name)
        if isinstance(val, str):
            super().set(addNS(name), val)
        elif isinstance(val, (int, float)):
            super().set(addNS(name), str(val))
        elif isinstance(val, list):
            super().set(addNS(name), ",".join(map(str, val)))
        elif isinstance(val, BaseElement):
            # Order of children:
            # Property Element(s) -> text -> child1 -> child1.tail -> c2 -> c2.tail
            prophandle = etree.Element(self.get_tag_name(name))
            self.insert(0, prophandle)
            if self.text != "":
                prophandle.tail = self.text
                self.text = ""
            prophandle.append(val)

    def remove_property(self, name):
        """Remove a property"""
        for node in self:
            if super(etree.ElementBase, node).tag == self.get_tag_name(name):
                self.remove(node)
        if addNS(name) in self.keys():
            self.attrib.pop(addNS(name))  # pylint: disable=no-member

    def get(self, name, default=None):
        if ":" not in name:
            for node in self.xpath("/" + self.get_tag_name(name)):
                return node
        return super().get(addNS(name), default)

    def add_property(
        self,
        name: str,
        val: Optional[Union[str, Number, XAMLObject, List[Number]]],
        default=None,
    ):
        """Add a property if not None"""
        if val is not None:
            if val == default:
                self.remove_property(name)
            else:
                self.set(name, val)

    def add_properties(
        self, properties: Dict[str, Union[str, XAMLObject, Number, List[Number]]]
    ):
        """Add multiple properties if not None"""
        for key, value in properties.items():
            self.add_property(key, value)

    def add_children(self, children: List[XAMLObject]):
        """Append multiple children"""
        for child in children:
            self.append(child)

    def add_name(self, name: str):
        """Adding names lead to an exception when parsing the XAML if it's wrapped in a
        resource dictionary."""

    @property
    def key(self):
        """Get the objects' x:Key attribute."""
        return self.get(KEY_ATTR, None)

    @property
    def root(self):
        """Get the root ResourceDictionary from any element descendent"""
        if self.getparent() is not None:
            return self.getparent().root
        if not isinstance(self, ResourceDictionary):
            raise ValueError(
                "Element fragment does not have a ResourceDictionary root!"
            )
        return self

    def add_child(self, element: Optional[XAMLObject], front=False):
        """Append/insert a child if not none"""
        if element is None:
            return
        if not front:
            self.append(element)
        else:
            # TODO append after property elements
            self.insert(0, element)

    @staticmethod
    def remove_key(element):
        """remove the x:key attribute if set"""
        if element.get(KEY_ATTR, None) is not None:
            element.attrib.pop(addNS(KEY_ATTR))

    def append(self, element):
        """Append without key (key is only valid inside ResourceDictionary)"""
        BaseElement.remove_key(element)
        self.make_key_unique(self, element, "Name", False)
        super().append(element)

    def insert(self, index, element):
        """Insert without key (key is only valid inside ResourceDictionary)"""
        BaseElement.remove_key(element)
        self.make_key_unique(self, element, "Name", False)
        super().insert(index, element)

    def xpath(self, expr, namespaces=None):
        return super().xpath(
            expr, namespaces=namespaces or GLOBAL_SETTINGS.xaml_namespaces
        )

    def parse_transform(self, attribute="Transform") -> inkex.Transform:
        """Reads the transform attribute and returns an inkex.Transform object.
        This currently does not support TransformGroup"""
        read = self.get(attribute, None)
        if read is None:
            return inkex.Transform()
        return inkex.Transform(f"matrix({read})")

    def __str__(self):
        return f"{self.TAG} element with {len(self.getchildren())} children"

    def set_invisible(self):
        """Makes an element invisible"""
        self.add_property("ClipGeometry", "")

    @staticmethod
    def transform_to_xaml(transform: inkex.Transform):
        """Converts an inkex Transform struct to an XAML object representation"""
        as_string = ",".join(map(str, transform.to_hexad()))
        if GLOBAL_SETTINGS.avalonia:
            result = MatrixTransform()
            result.set("Matrix", as_string)
            return result
        return as_string

    def set_transform(self, transform: inkex.Transform):
        """Sets a transform on an object"""
        if transform != inkex.Transform():  # don't add the identity transform
            self.add_property("Transform", self.transform_to_xaml(transform))


class BaseElement(XAMLObject):
    """Superclass for any XAML object"""


class ResourceDictionary(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.resourcedictionary"""

    def insert_wrapper(self, element: BaseElement, insert_function: Callable):
        """Inserting with unique key"""
        self.make_key_unique(self, element)
        key = element.get(KEY_ATTR, None)
        result = insert_function(element)
        element.set(addNS(KEY_ATTR), key)
        return result

    def append(self, element: BaseElement):
        """Appending with unique key"""
        return self.insert_wrapper(element, super().append)

    def insert(self, index: int, element: BaseElement):
        """Inserting with unique key"""
        inserter = super().insert
        return self.insert_wrapper(element, lambda el: inserter(index, el))

    def get_child(self, key):
        """Finds a direct subchild based on key"""
        for i in self.iterdescendants():
            if i.get(KEY_ATTR) == key:
                return i
        return None


class Drawing(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.drawing"""


class DrawingGroup(Drawing):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.drawinggroup"""


class GeometryDrawing(Drawing):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.geometrydrawing"""


class Geometry(Drawing):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.geometry"""


class PathGeometry(Geometry):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.pathgeometry"""


class LineGeometry(Geometry):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.linegeometry"""


class RectangleGeometry(Geometry):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.rectanglegeometry"""


class EllipseGeometry(Geometry):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.ellipsegeometry"""


class Brush(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.brush"""


class SolidColorBrush(Brush):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.solidcolorbrush"""


class DrawingBrush(Brush):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.drawingbrush"""


class GradientBrush(Brush):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.gradientbrush"""


class LinearGradientBrush(GradientBrush):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.lineargradientbrush"""


class RadialGradientBrush(GradientBrush):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.radialgradientbrush"""


class VisualBrush(Brush):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.visualbrush"""


class Pen(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.pen"""


class DashStyle(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.dashstyle"""


class GradientStop(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.gradientstop"""


class DrawingImage(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.drawingimage"""


class UIElement(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.controls.uielement"""

    def set_invisible(self):
        if GLOBAL_SETTINGS.wpf:
            self.add_property("Visibility", "Hidden")
        else:
            self.add_property("IsVisible", "False")

    def set_transform(self, transform: inkex.Transform):
        if transform != inkex.Transform():
            self.add_property("RenderTransform", self.transform_to_xaml(transform))
            if GLOBAL_SETTINGS.avalonia:
                self.add_property("RenderTransformOrigin", "0,0")


class Canvas(UIElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.controls.canvas"""


class Viewbox(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.controls.viewbox"""


class Shape(UIElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.shapes.shape"""


class Path(Shape):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.shapes.path"""


class Rectangle(Shape):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.shapes.rectangle"""


class Polygon(Shape):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.shapes.polygon"""


class Polyline(Shape):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.shapes.polyline"""


class Line(Shape):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.shapes.line"""


class Ellipse(Shape):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.shapes.ellipse"""


class GeometryGroup(Geometry):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.geometrygroup"""


class MatrixTransform(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.matrixtransform"""


class BlurEffect(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.media.effects.blureffect"""


class TextBlock(UIElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.controls.textblock"""


class Span(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.documents.span"""


class LineBreak(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.documents.linebreak"""


class TextDecoration(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.textdecoration"""


class TextDecorationCollection(BaseElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.textdecorationcollection"""


class Grid(UIElement):
    """Interface for
    https://docs.microsoft.com/dotnet/api/system.windows.controls.grid"""



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2006 Jos Hirth, kaioa.com
#               2020 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Example filltext sentences generated over at http://lipsum.com/
"""

import random

import inkex
from inkex import Layer, FlowRoot, FlowRegion, FlowPara, Rectangle, TextElement, Tspan

CORPA = [
    "Lorem ipsum dolor sit amet, consectetuer adipiscing elit. ",
    "Duis sem velit, ultrices et, fermentum auctor, rhoncus ut, ligula. ",
    "Phasellus at purus sed purus cursus iaculis. ",
    "Suspendisse fermentum. ",
    "Pellentesque et arcu. ",
    "Maecenas viverra. ",
    "In consectetuer, lorem eu lobortis egestas, velit odio imperdiet"
    " eros, sit amet sagittis nunc mi ac neque. ",
    "Sed non ipsum. ",
    "Nullam venenatis gravida orci. ",
    "Curabitur nunc ante, ullamcorper vel, auctor a, aliquam at, tortor. ",
    "Etiam sodales orci nec ligula. ",
    "Sed at turpis vitae velit euismod aliquet. ",
    "Fusce venenatis ligula in pede. ",
    "Pellentesque viverra dolor non nunc. ",
    "Donec interdum vestibulum libero. ",
    "Morbi volutpat. ",
    "Phasellus hendrerit. ",
    "Quisque dictum quam vel neque. ",
    "Quisque aliquam, nulla ac scelerisque convallis, nisi ligula sagittis"
    " risus, at nonummy arcu urna pulvinar nibh. ",
    "Nam pharetra. ",
    "Nam rhoncus, lectus vel hendrerit congue, nisl lorem feugiat ante, in"
    " fermentum erat nulla tristique arcu. ",
    "Mauris et dolor. ",
    "Vestibulum ante ipsum primis in faucibus orci luctus et ultrices posuere"
    " cubilia Curae; Donec gravida, ante vel ornare lacinia, orci enim porta"
    " est, eget sollicitudin lectus lectus eget lacus. ",
    "Praesent a lacus vitae turpis consequat semper. ",
    "In commodo, dolor quis fermentum ullamcorper, urna massa volutpat"
    " massa, vitae mattis purus arcu nec nulla. ",
    "In hac habitasse platea dictumst. ",
    "Praesent scelerisque. ",
    "Nullam sapien mauris, venenatis at, fermentum at, tempus eu, urna. ",
    "Vestibulum non arcu a ante feugiat vestibulum. ",
    "Nam laoreet dui sed magna. ",
    "Proin diam augue, semper vitae, varius et, viverra id, felis. ",
    "Pellentesque sit amet dui vel justo gravida auctor. ",
    "Aenean scelerisque metus eget sem. ",
    "Maecenas rhoncus rhoncus ipsum. ",
    "Donec nonummy lacinia leo. ",
    "Aenean turpis ipsum, rhoncus vitae, posuere vitae, euismod sed, ligula. ",
    "Pellentesque habitant morbi tristique senectus et netus et malesuada"
    " fames ac turpis egestas. ",
    "Mauris tempus diam. ",
    "Maecenas justo. ",
    "Sed a lorem ut est tincidunt consectetuer. ",
    "Ut eu metus id lectus vestibulum ultrices. ",
    "Suspendisse lectus. ",
    "Vivamus posuere, ante eu tempor dictum, felis nibh facilisis sem, eu"
    " auctor metus nulla non lorem. ",
    "Suspendisse potenti. ",
    "Integer fringilla. ",
    "Morbi urna. ",
    "Morbi pulvinar nulla sit amet nisl. ",
    "Mauris urna sem, suscipit vitae, dignissim id, ultrices sed, nunc. ",
    "Morbi a mauris. ",
    "Pellentesque suscipit accumsan massa. ",
    "Quisque arcu ante, cursus in, ornare quis, viverra ut, justo. ",
    "Quisque facilisis, urna sit amet pulvinar mollis, purus arcu adipiscing"
    " velit, non condimentum diam purus eu massa. ",
    "Suspendisse potenti. ",
    "Phasellus nisi metus, tempus sit amet, ultrices ac, porta nec, felis. ",
    "Aliquam metus. ",
    "Nam a nunc. ",
    "Vivamus feugiat. ",
    "Nunc metus. ",
    "Vestibulum ante ipsum primis in faucibus orci luctus et ultrices posuere"
    " cubilia Curae; Vivamus eu orci. ",
    "Sed elementum, felis quis porttitor sollicitudin, augue nulla sodales"
    " sapien, sit amet posuere quam purus at lacus. ",
    "Curabitur tincidunt tellus nec purus. ",
    "Nam consectetuer mollis dolor. ",
    "Sed quis elit. ",
    "Aenean luctus vulputate turpis. ",
    "Proin lectus orci, venenatis pharetra, egestas id, tincidunt vel, eros. ",
    "Nulla facilisi. ",
    "Aliquam vel nibh. ",
    "Vivamus nisi elit, nonummy id, facilisis non, blandit ac, dolor. ",
    "Etiam cursus purus interdum libero. ",
    "Nam id neque. ",
    "Etiam pede nunc, vestibulum vel, rutrum et, tincidunt eu, enim. ",
    "Aenean id purus. ",
    "Aenean ultrices turpis. ",
    "Mauris et pede. ",
    "Suspendisse potenti. ",
    "Aliquam velit dui, commodo quis, porttitor eget, convallis et, nisi. ",
    "Maecenas convallis dui. ",
    "In leo ante, venenatis eu, volutpat ut, imperdiet auctor, enim. ",
    "Mauris ac massa vestibulum nisl facilisis viverra. ",
    "Phasellus magna sem, vulputate eget, ornare sed, dignissim sit amet, pede. ",
    "Aenean justo ipsum, luctus ut, volutpat laoreet, vehicula in, libero. ",
    "Praesent semper, neque vel condimentum hendrerit, lectus elit pretium"
    "ligula, nec consequat nisl velit at dui. ",
    "Proin dolor sapien, adipiscing id, sagittis eu, molestie viverra, mauris. ",
    "Aenean ligula. ",
    "Vestibulum ante ipsum primis in faucibus orci luctus et ultrices posuere"
    " cubilia Curae; Suspendisse potenti. ",
    "Etiam pharetra lacus sed velit imperdiet bibendum. ",
    "Nunc in turpis ac lacus eleifend sagittis. ",
    "Nam massa turpis, nonummy et, consectetuer id, placerat ac, ante. ",
    "In tempus urna. ",
    "Quisque vehicula porttitor odio. ",
    "Aliquam sed erat. ",
    "Vestibulum viverra varius enim. ",
    "Donec ut purus. ",
    "Pellentesque convallis dolor vel libero. ",
    "Integer tempus malesuada pede. ",
    "Integer porta. ",
    "Donec diam eros, tristique sit amet, pretium vel, pellentesque ut, neque. ",
    "Nulla blandit justo a metus. ",
    "Curabitur accumsan felis in erat. ",
    "Curabitur lorem risus, sagittis vitae, accumsan a, iaculis id, metus. ",
    "Nulla sagittis condimentum ligula. ",
    "Aliquam imperdiet lobortis metus. ",
    "Suspendisse molestie sem. ",
    "Ut venenatis. ",
    "Pellentesque condimentum felis a sem. ",
    "Fusce nonummy commodo dui. ",
    "Nullam libero nunc, tristique eget, laoreet eu, sagittis id, ante. ",
    "Etiam fermentum. ",
    "Phasellus auctor enim eget sem. ",
    "Morbi turpis arcu, egestas congue, condimentum quis, tristique cursus, leo. ",
    "Sed fringilla. ",
    "Nam malesuada sapien eu nibh. ",
    "Pellentesque ac turpis. ",
    "Nulla sed lacus. ",
    "Mauris sed nulla quis nisi interdum tempor. ",
    "Quisque pretium rutrum ligula. ",
    "Mauris tempor ultrices justo. ",
    "In hac habitasse platea dictumst. ",
    "Donec sit amet enim. ",
    "Suspendisse venenatis. ",
    "Nam nisl quam, posuere non, volutpat sed, semper vitae, magna. ",
    "Donec ut urna. ",
    "Integer risus velit, facilisis eget, viverra et, venenatis id, leo. ",
    "Cras facilisis felis sit amet lorem. ",
    "Nam molestie nisl at metus. ",
    "Suspendisse viverra placerat tortor. ",
    "Phasellus lacinia iaculis mi. ",
    "Sed dolor. ",
    "Quisque malesuada nulla sed pede volutpat pulvinar. ",
    "Cras gravida. ",
    "Mauris tincidunt aliquam ante. ",
    "Fusce consectetuer tellus ut nisl. ",
    "Curabitur risus urna, placerat et, luctus pulvinar, auctor vel, orci. ",
    "Class aptent taciti sociosqu ad litora torquent per conubia nostra, per inceptos hymenaeos. ",
    "Praesent aliquet, neque pretium congue mattis, ipsum augue dignissim ante, ac"
    " pretium nisl lectus at magna. ",
    "Vivamus quis mi. ",
    "Nam sed nisl nec elit suscipit ullamcorper. ",
    "Donec tempus quam quis neque. ",
    "Donec rutrum venenatis dui. ",
    "Praesent a eros. ",
    "Aliquam justo lectus, iaculis a, auctor sed, congue in, nisl. ",
    "Etiam non neque ac mi vestibulum placerat. ",
    "Donec at diam a tellus dignissim vestibulum. ",
    "Integer accumsan. ",
    "Cras ac enim vel dui vestibulum suscipit. ",
    "Pellentesque tempor. ",
    "Praesent lacus. ",
]


class LoremIpsum(inkex.EffectExtension):
    """Generate text with psuedo latin content"""

    def add_arguments(self, pars):
        pars.add_argument(
            "--num", type=int, default=5, help="Number of paragraphs to generate"
        )
        pars.add_argument(
            "-c", "--sentencecount", type=int, default=16, help="Number of Sentences"
        )
        pars.add_argument("-f", "--fluctuation", type=int, default=4, help="+/-")
        pars.add_argument("--tab", help="The selected UI-tab when OK was pressed")
        pars.add_argument(
            "--svg2", help="Use SVG2 flowed text", default=True, type=inkex.Boolean
        )

    def make_paragraph(self, text_index=0):
        """Make a paragraph"""
        _min = max(1, self.options.sentencecount - self.options.fluctuation)
        _max = max(2, self.options.sentencecount + self.options.fluctuation)
        scount = int(random.random() * _max + _min)
        for sentence in range(scount):
            if text_index + sentence == 0:
                yield CORPA[0]
            else:
                index = int(random.random() * (len(CORPA) - 1))
                yield CORPA[index]

    def add_text_svg12(self, node):
        """Create many flowed text paragraph and append to node"""
        for text_index in range(self.options.num):
            para = node.add(FlowPara())
            para.text = "".join(self.make_paragraph(text_index))
            node.append(FlowPara())

    def add_text_svg2(self, node):
        """Add paragraphs to SVG2 flowed text node"""
        tspan = node.add(Tspan())
        newtext = "\n\n".join(
            [
                "".join(self.make_paragraph(text_index))
                for text_index in range(self.options.num)
            ]
        )
        tspan.text = newtext

    def get_layer(self):
        """Returns the current layer if set, otherwise creates a new layer to the document and
        returns it"""
        parent = self.svg.get_current_layer()
        if parent is None:
            parent = self.svg.add(Layer.new("lorum ipsum"))
        return parent

    def create_text_svg2(self, shape):
        """Creates a new SVG2 flowed text with the given shape inside. If no shape inside was set,
        the flowed text is appended to the selected layer"""
        style = inkex.Style()
        if shape is not None and not isinstance(shape, TextElement):
            parent = shape.getparent()
            style["shape-inside"] = shape.get_id(as_url=2)
        else:
            parent = self.get_layer()
            style["inline-size"] = self.svg.viewbox_width
        textelement = parent.add(TextElement())
        textelement.style = style
        textelement.style["white-space"] = "pre"
        textelement.style["font-size"] = self.svg.viewport_to_unit("8pt")
        self.add_text_svg2(textelement)

    def create_text_svg12(self, shape):
        """Creates a new SVG1.2 flowed text with the given shape inside. If no shape inside was set,
        the flowed text is appended to the selected layer"""
        root = FlowRoot()
        root.set("xml:space", "preserve")
        root.style["font-size"] = self.svg.viewport_to_unit("8pt")
        region = root.add(FlowRegion())
        if shape is not None and not isinstance(shape, TextElement):
            parent = shape.getparent()
            # For svg1.2 flowed text, store a copy of the shape inside the flowregion.
            region.add(shape.copy())
        else:
            # Nothing selected, create a new flowtext
            parent = self.get_layer()
            shape = region.add(
                Rectangle(
                    x="0",
                    y="0",
                    width=str(int(self.svg.viewbox_width)),
                    height=str(int(self.svg.viewbox_height)),
                )
            )
        parent.add(root)
        self.add_text_svg12(root)

    def effect(self):
        # Existing text flow to insert new text into
        done = False
        for node in self.svg.selection.filter(FlowRoot):
            self.add_text_svg12(node)
            done = True
        for node in self.svg.selection.filter(TextElement):
            shape = node.style.get("shape-inside")
            inlinesize = node.style.get("inline-size")
            if (
                shape is not None and self.svg.getElementById(shape[4:-1]) is not None
            ) or inlinesize is not None:
                self.add_text_svg2(node)
                done = True

        if done:
            return

        shape = self.svg.selection.first()
        if self.options.svg2:
            self.create_text_svg2(shape)
        else:
            self.create_text_svg12(shape)


if __name__ == "__main__":
    LoremIpsum().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2020-2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Test utilities for the XAML exporter"""

from io import BytesIO
import re

from lxml import etree

from inkex.tester import TestCase

from .svg2xaml import XamlOutput


class Svg2XamlTester(TestCase):
    """Base class for SVG->XAML conversion tests"""

    KEY_ATTR = "{http://schemas.microsoft.com/winfx/2006/xaml}Key"

    @staticmethod
    def cleanup_tree(subtree: etree.Element) -> etree.Element:
        """
        Sanitize the XAML tree for better comparability.
        - Put child-properties (i.e. Canvas.Clip or similar) last and alphabetically
        ordered
        """
        properties = []
        for child in list(subtree):
            # after namespace, there's a dot in the tag
            if "." in child.tag.split("}")[1]:
                properties += [child]
                subtree.remove(child)
        properties.sort(key=lambda x: x.tag)
        for prop in properties:
            subtree.append(prop)
        for child in subtree:
            Svg2XamlTester.cleanup_tree(child)
        return subtree

    @staticmethod
    def run_to_tree(svg, args):
        exporter = XamlOutput()
        out = BytesIO()
        exporter.run(args + [svg], output=out)
        out.seek(0)
        decoded = out.read().decode("utf-8")
        return Svg2XamlTester.cleanup_tree(etree.fromstring(decoded))

    xamlns = "http://schemas.microsoft.com/winfx/2006/xaml/presentation"
    namespaces = {
        "x": "http://schemas.microsoft.com/winfx/2006/xaml",
        "d": "http://schemas.microsoft.com/expression/blend/2008",
        "xaml": xamlns,
    }

    def run_to_dg(self, svg, args=None) -> etree.Element:
        """Runs the extensions, and fetches the toplevel DrawingGroup."""
        elements = Svg2XamlTester.run_to_tree(
            self.data_file("svg", svg), args or ["--target=wpf", "--mode=lowlevel"]
        ).xpath("//xaml:DrawingGroup", namespaces=self.namespaces)
        return elements[0]

    def run_to_canvas(self, svg, args=None) -> etree.Element:
        """Runs the extensions, and fetches the toplevel Canvas."""
        elements = Svg2XamlTester.run_to_tree(
            self.data_file("svg", svg), args or ["--target=wpf", "--mode=canvas"]
        ).xpath("//xaml:Canvas", namespaces=self.namespaces)
        return elements[0]

    def assertTagEqual(self, item, tag):
        """Check if the tag of item is "tag" """
        self.assertEqual(item.tag, f"{{{self.xamlns}}}{tag}")

    def assertPropertyEqual(
        self, item, prop, expected_value, default=None, places=10, separator=",| "
    ):
        """Datatype-aware comparison of properties"""

        value = item.get(prop, default)
        if expected_value is None:
            self.assertIsNone(value)
            return
        # Numbers
        if prop in [
            "Canvas.Left",
            "Canvas.Top",
            "FontSize",
            "Height",
            "LineHeight",
            "Offset",
            "PenOffset",
            "RadiusX",
            "RadiusY",
            "Radius",
            "StrokeThickness",
            "Thickness",
            "Width",
            "X1",
            "X2",
            "Y1",
            "Y2",
        ]:
            self.assertAlmostEqual(float(value), expected_value, places=places)

        # Lists of numbers
        elif prop in [
            "Center",
            "Dashes",
            "DestinationRect",
            "EndPoint",
            "GradientOrigin",
            "Matrix",
            "Rect",
            "RenderTransform",
            "RenderTransformOrigin",
            "SourceRect",
            "StartPoint",
            "StrokeDashArray",
            "Transform",
            "Viewbox",
            "Viewport",
        ]:
            self.assertAlmostTuple(
                [float(i) for i in re.split(separator, value) if i != ""],
                expected_value,
                precision=places,
            )

        # For other data types, compare literally
        else:
            self.assertEqual(value, expected_value)



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2008-2009 Alvin Penner, penner@vaxxine.com
#               2009, Christian Mayer, inkscape@christianmayer.de
#               2020, MartinOwens, doctormo@geek-2.com
#               2024, Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Input a DXF file >= (AutoCAD Release 13 == AC1012)
"""

import os
import re
import struct
import sys
import math

from collections import defaultdict
from typing import Tuple
from urllib.parse import quote
from lxml import etree

import inkex
from inkex.localization import inkex_gettext as _

global defs
global block  # 2021.6
global layer
global svg
global scale
global xmin
global ymin
global height
global style_font3
global style_direction

COLORS = [
    "PAD",
    "#FF0000",
    "#FFFF00",
    "#00FF00",
    "#00FFFF",
    "#0000FF",
    "#FF00FF",
    "#000000",
    "#808080",
    "#C0C0C0",
    "#FF0000",
    "#FF7F7F",
    "#CC0000",
    "#CC6666",
    "#990000",
    "#994C4C",
    "#7F0000",
    "#7F3F3F",
    "#4C0000",
    "#4C2626",
    "#FF3F00",
    "#FF9F7F",
    "#CC3300",
    "#CC7F66",
    "#992600",
    "#995F4C",
    "#7F1F00",
    "#7F4F3F",
    "#4C1300",
    "#4C2F26",
    "#FF7F00",
    "#FFBF7F",
    "#CC6600",
    "#CC9966",
    "#994C00",
    "#99724C",
    "#7F3F00",
    "#7F5F3F",
    "#4C2600",
    "#4C3926",
    "#FFBF00",
    "#FFDF7F",
    "#CC9900",
    "#CCB266",
    "#997200",
    "#99854C",
    "#7F5F00",
    "#7F6F3F",
    "#4C3900",
    "#4C4226",
    "#FFFF00",
    "#FFFF7F",
    "#CCCC00",
    "#CCCC66",
    "#989800",
    "#98984C",
    "#7F7F00",
    "#7F7F3F",
    "#4C4C00",
    "#4C4C26",
    "#BFFF00",
    "#DFFF7F",
    "#99CC00",
    "#B2CC66",
    "#729800",
    "#85984C",
    "#5F7F00",
    "#6F7F3F",
    "#394C00",
    "#424C26",
    "#7FFF00",
    "#BFFF7F",
    "#66CC00",
    "#99CC66",
    "#4C9800",
    "#72984C",
    "#3F7F00",
    "#5F7F3F",
    "#264C00",
    "#394C26",
    "#3FFF00",
    "#9FFF7F",
    "#33CC00",
    "#7FCC66",
    "#269800",
    "#5F984C",
    "#1F7F00",
    "#4F7F3F",
    "#134C00",
    "#2F4C26",
    "#00FF00",
    "#7FFF7F",
    "#00CC00",
    "#66CC66",
    "#009800",
    "#4C984C",
    "#007F00",
    "#3F7F3F",
    "#004C00",
    "#264C26",
    "#00FF3F",
    "#7FFF9F",
    "#00CC33",
    "#66CC7F",
    "#009826",
    "#4C985F",
    "#007F1F",
    "#3F7F4F",
    "#004C13",
    "#264C2F",
    "#00FF7F",
    "#7FFFBF",
    "#00CC66",
    "#66CC99",
    "#00984C",
    "#4C9872",
    "#007F3F",
    "#3F7F5F",
    "#004C26",
    "#264C39",
    "#00FFBF",
    "#7FFFDF",
    "#00CC99",
    "#66CCB2",
    "#009872",
    "#4C9885",
    "#007F5F",
    "#3F7F6F",
    "#004C39",
    "#264C42",
    "#00FFFF",
    "#7FFFFF",
    "#00CCCC",
    "#66CCCC",
    "#009898",
    "#4C9898",
    "#007F7F",
    "#3F7F7F",
    "#004C4C",
    "#264C4C",
    "#00BFFF",
    "#7FDFFF",
    "#0099CC",
    "#66B2CC",
    "#007298",
    "#4C8598",
    "#005F7F",
    "#3F6F7F",
    "#00394C",
    "#26424C",
    "#007FFF",
    "#7FBFFF",
    "#0066CC",
    "#6699CC",
    "#004C98",
    "#4C7298",
    "#003F7F",
    "#3F5F7F",
    "#00264C",
    "#26394C",
    "#003FFF",
    "#7F9FFF",
    "#0033CC",
    "#667FCC",
    "#002698",
    "#4C5F98",
    "#001F7F",
    "#3F4F7F",
    "#00134C",
    "#262F4C",
    "#0000FF",
    "#7F7FFF",
    "#0000CC",
    "#6666CC",
    "#000098",
    "#4C4C98",
    "#00007F",
    "#3F3F7F",
    "#00004C",
    "#26264C",
    "#3F00FF",
    "#9F7FFF",
    "#3300CC",
    "#7F66CC",
    "#260098",
    "#5F4C98",
    "#1F007F",
    "#4F3F7F",
    "#13004C",
    "#2F264C",
    "#7F00FF",
    "#BF7FFF",
    "#6600CC",
    "#9966CC",
    "#4C0098",
    "#724C98",
    "#3F007F",
    "#5F3F7F",
    "#26004C",
    "#39264C",
    "#BF00FF",
    "#DF7FFF",
    "#9900CC",
    "#B266CC",
    "#720098",
    "#854C98",
    "#5F007F",
    "#6F3F7F",
    "#39004C",
    "#42264C",
    "#FF00FF",
    "#FF7FFF",
    "#CC00CC",
    "#CC66CC",
    "#980098",
    "#984C98",
    "#7F007F",
    "#7F3F7F",
    "#4C004C",
    "#4C264C",
    "#FF00BF",
    "#FF7FDF",
    "#CC0099",
    "#CC66B2",
    "#980072",
    "#984C85",
    "#7F005F",
    "#7F3F6F",
    "#4C0039",
    "#4C2642",
    "#FF007F",
    "#FF7FBF",
    "#CC0066",
    "#CC6699",
    "#98004C",
    "#984C72",
    "#7F003F",
    "#7F3F5F",
    "#4C0026",
    "#4C2639",
    "#FF003F",
    "#FF7F9F",
    "#CC0033",
    "#CC667F",
    "#980026",
    "#984C5F",
    "#7F001F",
    "#7F3F4F",
    "#4C0013",
    "#4C262F",
    "#333333",
    "#5B5B5B",
    "#848484",
    "#ADADAD",
    "#D6D6D6",
    "#FFFFFF",
]


def get_rgbcolor(dxfcolor, parent_color="#000000"):
    """Returns hex color code corresponding to a color value

    dxfcolor     -- dxf code to convert to hex color code
                    0 (BYBLOCK) and 256 (BYLAYER) use parent_color
                    No more differentiation is currently done
                    Negative values are ignored (specification
                    allows layer to be hidden here)
                    Negative values also use parent_color
    parent_color -- hex color code from parent layer.
                    Use default color '#000000' if
                    parent layer color undefined.
    """
    rgbcolor = None
    if dxfcolor in range(1, len(COLORS)):
        rgbcolor = COLORS[dxfcolor]
    if not rgbcolor:
        rgbcolor = parent_color
    return rgbcolor


class ValueConstruct(defaultdict):
    """Store values from the DXF and provide them as named attributes"""

    values = {
        "1": ("text", "default"),
        "2": ("tag", "block_name"),
        "3": ("mtext",),
        "6": ("line_type",),
        "7": ("text_style",),
        "8": ("layer_name",),
        "10": ("x1",),
        "11": ("x2",),
        "13": ("x3",),
        "14": ("x4",),
        "20": ("y1",),
        "21": ("y2",),
        "23": ("y3",),
        "24": ("y4",),
        "40": ("scale", "knots", "radius", "width_ratio"),
        "41": ("ellipse_a1", "insert_scale_x", "mtext_width"),
        "42": ("ellipse_a2", "bulge", "insert_scale_y"),
        "50": ("angle",),
        "51": ("angle2",),
        "62": ("color",),
        "70": ("fill", "flags"),
        "71": ("attach_pos",),
        "72": ("edge_type",),
        "73": ("sweep",),  # ccw
        "92": ("path_type",),
        "93": ("num_edges",),
        "230": ("extrude",),
        "370": ("line_weight",),
    }
    attrs = dict([(name, a) for a, b in values.items() for name in b])

    def __init__(self):
        super().__init__(list)

    @classmethod
    def is_valid(cls, key):
        return key in cls.values

    def __getattr__(self, attr):
        is_list = attr.endswith("_list")
        key = attr[:-5] if is_list else attr
        if key in self.attrs:
            ret = self[self.attrs[key]]
            if not attr.endswith("_list"):
                return ret[0]
            return ret
        if attr.startswith("has_"):
            key = attr[4:]
            if key in self.attrs:
                return self.attrs[key] in self
        raise AttributeError(f"Can't find dxf attribute '{key}' {attr}")

    def __setattr__(self, attr, value):
        if not attr in self.attrs:
            raise AttributeError(f"Can't set bad dxf attribute '{attr}'")
        if not isinstance(value, list):
            value = [value]
        self[self.attrs[attr]] = value

    def adjust_coords(self, xmin, ymin, scale, extrude, height):
        """Adjust the x,y coordinates to fit on the page"""
        for xgrp in set(["10", "11", "13", "14"]) & set(self):  # scale/reflect x values
            for i in range(len(self[xgrp])):
                self[xgrp][i] = scale * (extrude * self[xgrp][i] - xmin)
        for ygrp in set(["20", "21", "23", "24"]) & set(self):  # scale y values
            for i in range(len(self[ygrp])):
                self[ygrp][i] = height - scale * (self[ygrp][i] - ymin)


export_viewport = False
export_endsec = False


def re_hex2unichar(m):
    # return unichr(int(m.group(1), 16))
    return chr(int(m.group(1), 16))


def formatStyle(style):
    return str(inkex.Style(style))


def export_text(vals):
    # mandatory group codes : (11, 12, 72, 73) (fit_x, fit_y, horizon, vertical)
    # TODO: position to display at by (x2,y2) according to 72(horizon),73(vertical)
    # groupcode 72:0(left),1(center),2(right),3(both side),4(middle),5(fit)
    # grouocode 73:0(standard),1(floor),2(center),3(ceiling)
    vals["71"].append(1)  # attach=pos left in mtext
    vals["70"].append(1)  # text: flags=1
    return export_mtext(vals)


def export_mtext(vals):
    # mandatory group codes : (1 or 3, 10, 20) (text, x, y)
    # TODO: text-format: \Font; \W; \Q; \L..\l etc
    if (vals.has_text or vals.has_mtext) and vals.has_x1 and vals.has_y1:
        x = vals.x1
        y = vals.y1
        # optional group codes : (21, 40, 50) (direction, text height mm, text angle)
        # optional group codes : 2: char style is defined at TABLES Section
        size = 12  # default fontsize in px
        if vals.has_scale:
            size = scale * textscale * vals.scale

        dx = dy = 0
        if not vals.has_flags:  # as mtext, putting in the box
            dy = size

        anchor = "start"
        if vals.has_attach_pos:
            if vals.attach_pos in (2, 5, 8):
                anchor = "middle"
            elif vals.attach_pos in (3, 6, 9):
                anchor = "end"
            if vals.attach_pos in (4, 5, 6):
                dy = size / 2
        # if vals.has_text_style and vals.text_style in style_font3 :
        #    attribs = {'x': '%f' % x, 'y': '%f' % y, 'style': 'font-size: %.3fpx; fill: %s; font-family: %s' \
        #            % (size, color, style_font3[vals.text_style])}
        # else :
        #    attribs = {'x': '%f' % x, 'y': '%f' % y, 'style': 'font-size: %.3fpx; fill: %s; font-family: %s' % (size, color, options.font)}
        attribs = {
            "x": "%f" % x,
            "y": "%f" % y,
            "style": "font-size: %.3fpx; fill: %s; font-family: %s; text-anchor: %s"
            % (size, color, options.font, anchor),
        }

        angle = 0  # default angle in degrees
        bVertical = False
        if vals.has_angle:  # TEXT only
            if vals.angle != 0:
                angle = vals.angle
            # attribs.update({'transform': 'rotate (%f %f %f)' % (-angle, x, y)})
        elif vals.has_y2 and vals.has_x2:
            # MTEXT
            # recover original data
            # (x,y)=(scale*(x-xmin), height-scale*(y-ymin)
            orgx = vals.x2 / scale + xmin
            orgy = -(vals.y2 - height) / scale + ymin
            unit = math.sqrt(orgy * orgy + orgx * orgx)
            if (unit < 1.01) and (unit > 0.99):
                ang1 = math.atan2(orgy, orgx)
                angle = 180 * ang1 / math.pi
            # attribs.update({'transform': 'rotate (%f %f %f)' % (-angle, x, y)})

        if vals.has_text_style and vals.text_style in style_direction:
            if style_direction[vals.text_style] & 4:
                # angle = -90
                # attribs.update({'transform': 'rotate (%f %f %f)' % (-angle, x, y)})
                bVertical = True
                angle = 0
                dx = size
                attribs = {
                    "x": "%f" % x,
                    "y": "%f" % y,
                    "style": "font-size: %.3fpx; fill: %s; font-family: %s; text-anchor: %s; writing-mode: tb"
                    % (size, color, options.font, anchor),
                }
        if angle != 0:
            attribs.update({"transform": "rotate (%f %f %f)" % (-angle, x, y)})

        node = layer.add(inkex.TextElement(**attribs))
        node.set("sodipodi:linespacing", "125%")
        text = ""
        if vals.has_mtext:
            text = "".join(vals.mtext_list)
        if vals.has_text:
            text += vals.text
        # folding long text
        if vals.has_mtext_width and vals.has_scale:
            if vals.mtext_width > 10.0 and vals.scale > 1.0:
                charsperline = vals.mtext_width * 2 / vals.scale  # or 1.6?
                nochars = int(charsperline)
                if len(text) > charsperline:
                    # plain text only no \P,{} better divide by space
                    if (text.find(r"\P") < 0) and (text.find(r"{") < 0):
                        pos = 0
                        while len(text) > pos:
                            text = text[:pos] + r"\P" + text[pos:]
                            pos += nochars + 2

        text = mtext_normalize(text)
        lines = 0
        found = text.find(r"\P")  # new line
        while found > -1:
            tspan = node.add(inkex.Tspan())
            if bVertical:
                tspan.set("y", "%f" % y)
                if lines > 0:
                    tspan.set("dx", "%f" % size)
            else:
                tspan.set("sodipodi:role", "line")
                if lines > 0:
                    tspan.set("x", x + dx)
                else:
                    tspan.set("dx", "%f" % dx)
                tspan.set("dy", "%f" % dy)
            # tspan.text = text[:found]
            text1 = text[:found]
            mtext_separate(node, tspan, text1)
            text = text[(found + 2) :]
            found = text.find(r"\P")
            lines += 1

        tspan = node.add(inkex.Tspan())
        if bVertical:
            tspan.set("y", "%f" % y)
            if lines > 0:
                tspan.set("dx", "%f" % dx)
        else:
            tspan.set("sodipodi:role", "line")
            if lines > 0:
                tspan.set("x", x + dx)
            else:
                tspan.set("dx", "%f" % dx)
            tspan.set("dy", "%f" % dy)
        # tspan.text = text
        text1 = text
        mtext_separate(node, tspan, text1)


def mtext_normalize(text):
    # {  \P  } -> {  }\P{  }
    found = text.find(r"\P")
    while found > -1:
        nest = False
        posL = text.rfind(r"{", 0, found)
        posR = text.rfind(r"}", 0, found)
        if posL > -1:
            if posR == -1:
                nest = True
            else:
                if posL > posR:
                    nest = True
        if nest:
            # paste }\P{
            control = ""
            if text[posL + 1] == "\\":
                posC = text.find(r";", posL)
                if posC != -1 and (posC - posL) < 20:
                    control = text[posL + 1 : posC + 1]
            text = text[:found] + r"}\P{" + control + text[found + 2 :]
        found = text.find(r"\P", found + 2)
    return text


def mtext_separate(node, tspan, text):
    # sparate aaa{bbb}(ccc) -> aaa,bbb.ccc
    tspanAdd = True
    found = text.find(r"{")
    while found > -1:
        if found == 0:
            found1 = text.find(r"}")
            if found1 < 1:
                break
            text1 = text[:found1]  # tspan
            text1 = text1[found + 1 :]
            if tspanAdd == False:
                tspan = node.add(inkex.Tspan())
            mtext_ctrl(tspan, text1)
            # tspan.text = text1 +'+1'
            tspanAdd = False
            text = text[found1 + 1 :]
            found = text.find(r"{")
        else:
            text1 = text[:found]  # tspan
            if tspanAdd == False:
                tspan = node.add(inkex.Tspan())
            mtext_ctrl(tspan, text1)
            # tspan.text = text1 +'+2'
            tspanAdd = False
            text = text[found:]
            found = 0

    if len(text) > 0:
        text1 = text
        if tspanAdd == False:
            tspan = node.add(inkex.Tspan())
        mtext_ctrl(tspan, text1)
        # tspan.text = text1 +'+3'
        tspanAdd = False


def mtext_ctrl(tspan, phrase):
    if len(phrase) == 0:
        return
    if phrase[0] != "\\":
        tspan.text = phrase
        return
    # if you'll add the function, you should remove the auto re.sub at setting group code:1
    if phrase[1].upper() in ("C", "H", "T", "Q", "W", "A"):
        # get the value
        found = phrase.find(r";")
        if found > 2:
            cvalue = phrase[:found]
            cvalue = cvalue[2:]
            try:
                value = float(cvalue)
            except ValueError:
                done = False
            else:
                done = True
                if phrase[1].upper() == "C":
                    i = int(value)
                    color = get_rgbcolor(i)
                    tspan.set("style", "stroke: %s" % color)
                elif phrase[1].upper() == "H":
                    value *= scale
                    tspan.set("style", "font-size: %.3fpx;" % value)
                elif phrase[1].upper() == "T":
                    tspan.set("style", "letter-spacing: %f;" % value)
                elif phrase[1].upper() == "A":
                    if value == 0:
                        tspan.set("dominant-baseline", "text-bottom")
                    elif value == 1:
                        tspan.set("dominant-baseline", "central")
                    elif value == 2:
                        tspan.set("dominant-baseline", "text-top")
                tspan.text = phrase[found + 1 :]
        else:
            tspan.text = phrase
    else:
        if phrase[1].upper() == "F":
            # get the value font-name & style & cut from text             2022.March
            # \FArial|b0|i0|c0|p0; b:bold,i:italic,c:charset,ppitch
            found = phrase.find(r";")
            if found > 2:
                cvalue = phrase[:found]
                tspan.text = phrase[found + 1 :]
        else:
            tspan.text = phrase


def export_point(vals, w):
    # mandatory group codes : (10, 20) (x, y)
    if vals.has_x1 and vals.has_y1:
        if vals["70"]:
            inkex.errormsg("$PDMODE is ignored. A point is displayed as normal.")
        if options.gcodetoolspoints:
            generate_gcodetools_point(vals.x1, vals.y1)
        else:
            generate_ellipse(vals.x1, vals.y1, w / 2, 0.0, 1.0, 0.0, 0.0)


def export_line(vals):
    """Draw a straight line from the dxf"""
    # mandatory group codes : (10, 11, 20, 21) (x1, x2, y1, y2)
    if vals.has_x1 and vals.has_x2 and vals.has_y1 and vals.has_y2:
        path = inkex.PathElement()
        path.style = style
        path.path = "M %f,%f %f,%f" % (vals.x1, vals.y1, vals.x2, vals.y2)
        layer.add(path)


def export_solid(vals):
    # arrows of dimension
    # mandatory group codes : (10, 11, 12, 20, 21, 22) (x1, x2, x3, y1, y2, y3)
    # TODO: 4th point
    if (
        vals.has_x1
        and vals.has_x2
        and vals.has_x3
        and vals.has_y1
        and vals.has_y2
        and vals.has_y3
    ):
        path = inkex.PathElement()
        path.style = style
        path.path = "M %f,%f %f,%f %f,%f z" % (
            vals.x1,
            vals.y1,
            vals.x2,
            vals.y2,
            vals.x3,
            vals.y3,
        )
        layer.add(path)


def export_spline(vals):
    # see : http://www.mactech.com/articles/develop/issue_25/schneider.html
    # mandatory group codes : (10, 20, 40, 70) (x[], y[], knots[], flags)
    if (
        vals.has_flags
        and vals.has_knots
        and vals.x1_list
        and len(vals.x1_list) == len(vals.y1_list)
    ):
        knots = vals.knots_list
        ctrls = len(vals.x1_list)
        if ctrls > 3 and len(knots) == ctrls + 4:  # cubic
            if ctrls > 4:
                for i in range(len(knots) - 5, 3, -1):
                    if knots[i] != knots[i - 1] and knots[i] != knots[i + 1]:
                        a0 = (knots[i] - knots[i - 2]) / (knots[i + 1] - knots[i - 2])
                        a1 = (knots[i] - knots[i - 1]) / (knots[i + 2] - knots[i - 1])
                        vals.x1_list.insert(
                            i - 1,
                            (1.0 - a1) * vals.x1_list[i - 2] + a1 * vals.x1_list[i - 1],
                        )
                        vals.y1_list.insert(
                            i - 1,
                            (1.0 - a1) * vals.y1_list[i - 2] + a1 * vals.y1_list[i - 1],
                        )
                        vals.x1_list[i - 2] = (1.0 - a0) * vals.x1_list[
                            i - 3
                        ] + a0 * vals.x1_list[i - 2]
                        vals.y1_list[i - 2] = (1.0 - a0) * vals.y1_list[
                            i - 3
                        ] + a0 * vals.y1_list[i - 2]
                        knots.insert(i, knots[i])
                for i in range(len(knots) - 6, 3, -2):
                    if (
                        knots[i] != knots[i + 2]
                        and knots[i - 1] != knots[i + 1]
                        and knots[i - 2] != knots[i]
                    ):
                        a1 = (knots[i] - knots[i - 1]) / (knots[i + 2] - knots[i - 1])
                        vals.x1_list.insert(
                            i - 1,
                            (1.0 - a1) * vals.x1_list[i - 2] + a1 * vals.x1_list[i - 1],
                        )
                        vals.y1_list.insert(
                            i - 1,
                            (1.0 - a1) * vals.y1_list[i - 2] + a1 * vals.y1_list[i - 1],
                        )
            ctrls = len(vals.x1_list)
            path = "M %f,%f" % (vals.x1, vals.y1)
            for i in range(0, (ctrls - 1) // 3):
                path += " C %f,%f %f,%f %f,%f" % (
                    vals.x1_list[3 * i + 1],
                    vals.y1_list[3 * i + 1],
                    vals.x1_list[3 * i + 2],
                    vals.y1_list[3 * i + 2],
                    vals.x1_list[3 * i + 3],
                    vals.y1_list[3 * i + 3],
                )
            if vals.flags & 1:  # closed path
                path += " z"
            attribs = {"d": path, "style": style}
            etree.SubElement(layer, "path", attribs)
        if ctrls == 3 and len(knots) == 6:  # quadratic
            path = "M %f,%f Q %f,%f %f,%f" % (
                vals.x1,
                vals.y1,
                vals.x1_list[1],
                vals.y1_list[1],
                vals.x1_list[2],
                vals.y1_list[2],
            )
            attribs = {"d": path, "style": style}
            etree.SubElement(layer, "path", attribs)
        if ctrls == 5 and len(knots) == 8:  # spliced quadratic
            path = "M %f,%f Q %f,%f %f,%f Q %f,%f %f,%f" % (
                vals.x1,
                vals.y1,
                vals.x1_list[1],
                vals.y1_list[1],
                vals.x1_list[2],
                vals.y1_list[2],
                vals.x1_list[3],
                vals.y1_list[3],
                vals.x1_list[4],
                vals.y1_list[4],
            )
            attribs = {"d": path, "style": style}
            etree.SubElement(layer, "path", attribs)


def export_circle(vals):
    # mandatory group codes : (10, 20, 40) (x, y, radius)
    if vals.has_x1 and vals.has_y1 and vals.has_radius:
        generate_ellipse(vals.x1, vals.y1, scale * vals.radius, 0.0, 1.0, 0.0, 0.0)


def export_arc(vals):
    # mandatory group codes : (10, 20, 40, 50, 51) (x, y, radius, angle1, angle2)
    if (
        vals.has_x1
        and vals.has_y1
        and vals.has_radius
        and vals.has_angle
        and vals.has_angle2
    ):
        generate_ellipse(
            vals.x1,
            vals.y1,
            scale * vals.radius,
            0.0,
            1.0,
            vals.angle * math.pi / 180.0,
            vals.angle2 * math.pi / 180.0,
        )


def export_ellipse(vals):
    # mandatory group codes : (10, 11, 20, 21, 40, 41, 42) (xc, xm, yc, ym, width ratio, angle1, angle2)
    if (
        vals.has_x1
        and vals.has_x2
        and vals.has_y1
        and vals.has_y2
        and vals.has_width_ratio
        and vals.has_ellipse_a1
        and vals.has_ellipse_a2
    ):
        # generate_ellipse(vals.x1, vals.y1, scale*vals.x2, scale*vals.y2, vals.width_ratio, vals.ellipse_a1, vals.ellipse_a2)
        # vals are through adjust_coords : recover proper value
        # (x,y)=(scale*x-xmin, height-scale*y-ymin)
        x2 = vals.x2 + xmin
        y2 = -vals.y2 + ymin + height
        generate_ellipse(
            vals.x1, vals.y1, x2, y2, vals.width_ratio, vals.ellipse_a1, vals.ellipse_a2
        )


def export_leader(vals):
    # mandatory group codes : (10, 20) (x, y)
    if vals.has_x1 and vals.has_y1:
        if len(vals.x1_list) > 1 and len(vals.y1_list) == len(vals.x1_list):
            path = "M %f,%f" % (vals.x1, vals.y1)
            for i in range(1, len(vals.x1_list)):
                path += " %f,%f" % (vals.x1_list[i], vals.y1_list[i])
            attribs = {"d": path, "style": style}
            etree.SubElement(layer, "path", attribs)


def export_polyline(vals):
    return export_lwpolyline(vals)


def export_lwpolyline(vals):
    # mandatory group codes : (10, 20, 70) (x, y, flags)
    if vals.has_x1 and vals.has_y1 and vals.has_flags:
        if len(vals.x1_list) > 1 and len(vals.y1_list) == len(vals.x1_list):
            # optional group codes : (42) (bulge)
            iseqs = 0
            ibulge = 0
            if vals.flags & 1:  # closed path
                seqs.append("20")
                vals.x1_list.append(vals.x1)
                vals.y1_list.append(vals.y1)
            while seqs[iseqs] != "20":
                iseqs += 1
            path = "M %f,%f" % (vals.x1, vals.y1)
            xold = vals.x1
            yold = vals.y1
            for i in range(1, len(vals.x1_list)):
                bulge = 0
                iseqs += 1
                while seqs[iseqs] != "20":
                    if seqs[iseqs] == "42":
                        bulge = vals.bulge_list[ibulge]
                        ibulge += 1
                    iseqs += 1
                if bulge:
                    sweep = 0  # sweep CCW
                    if bulge < 0:
                        sweep = 1  # sweep CW
                        bulge = -bulge
                    large = 0  # large-arc-flag
                    if bulge > 1:
                        large = 1
                    r = math.sqrt(
                        (vals.x1_list[i] - xold) ** 2 + (vals.y1_list[i] - yold) ** 2
                    )
                    r = 0.25 * r * (bulge + 1.0 / bulge)
                    path += " A %f,%f 0.0 %d %d %f,%f" % (
                        r,
                        r,
                        large,
                        sweep,
                        vals.x1_list[i],
                        vals.y1_list[i],
                    )
                else:
                    path += " L %f,%f" % (vals.x1_list[i], vals.y1_list[i])
                xold = vals.x1_list[i]
                yold = vals.y1_list[i]
            if vals.flags & 1:  # closed path
                path += " z"
            attribs = {"d": path, "style": style}
            etree.SubElement(layer, "path", attribs)


def export_hatch(vals):
    # mandatory group codes : (10, 20, 70, 72, 92, 93) (x, y, fill, Edge Type, Path Type, Number of edges)
    # TODO: Hatching Pattern
    if (
        vals.has_x1
        and vals.has_y1
        and vals.has_fill
        and vals.has_edge_type
        and vals.has_path_type
        and vals.has_num_edges
    ):
        if len(vals.x1_list) > 1 and len(vals.y1_list) == len(vals.x1_list):
            # optional group codes : (11, 21, 40, 50, 51, 73) (x, y, r, angle1, angle2, CCW)
            i10 = 1  # count start points
            i11 = 0  # count line end points
            i40 = 0  # count circles
            i72 = 0  # count edge type flags
            path = ""
            for i in range(0, len(vals.num_edges_list)):
                xc = vals.x1_list[i10]
                yc = vals.y1_list[i10]
                if vals.edge_type_list[i72] == 2:  # arc
                    rm = scale * vals.radius_list[i40]
                    a1 = vals.angle_list[i40]
                    path += "M %f,%f " % (
                        xc + rm * math.cos(a1 * math.pi / 180.0),
                        yc + rm * math.sin(a1 * math.pi / 180.0),
                    )
                else:
                    a1 = 0
                    path += "M %f,%f " % (xc, yc)
                for j in range(0, vals.num_edges_list[i]):
                    if vals.path_type_list[i] & 2:  # polyline
                        if j > 0:
                            path += "L %f,%f " % (vals.x1_list[i10], vals.y1_list[i10])
                        if j == vals.path_type_list[i] - 1:
                            i72 += 1
                    elif vals.edge_type_list[i72] == 2:  # arc
                        xc = vals.x1_list[i10]
                        yc = vals.y1_list[i10]
                        rm = scale * vals.radius_list[i40]
                        a2 = vals.angle2_list[i40]
                        diff = (a2 - a1 + 360) % 360
                        sweep = 1 - vals.sweep_list[i40]  # sweep CCW
                        large = 0  # large-arc-flag
                        if diff:
                            path += "A %f,%f 0.0 %d %d %f,%f " % (
                                rm,
                                rm,
                                large,
                                sweep,
                                xc + rm * math.cos(a2 * math.pi / 180.0),
                                yc + rm * math.sin(a2 * math.pi / 180.0),
                            )
                        else:
                            path += "A %f,%f 0.0 %d %d %f,%f " % (
                                rm,
                                rm,
                                large,
                                sweep,
                                xc + rm * math.cos((a1 + 180.0) * math.pi / 180.0),
                                yc + rm * math.sin((a1 + 180.0) * math.pi / 180.0),
                            )
                            path += "A %f,%f 0.0 %d %d %f,%f " % (
                                rm,
                                rm,
                                large,
                                sweep,
                                xc + rm * math.cos(a1 * math.pi / 180.0),
                                yc + rm * math.sin(a1 * math.pi / 180.0),
                            )
                        i40 += 1
                        i72 += 1
                    elif vals.edge_type_list[i72] == 1:  # line
                        path += "L %f,%f " % (vals.x2_list[i11], vals.y2_list[i11])
                        i11 += 1
                        i72 += 1
                    i10 += 1
                path += "z "
            if vals.has_fill:
                style = formatStyle({"fill": "%s" % color})
            else:
                style = formatStyle({"fill": "url(#Hatch)", "fill-opacity": "1.0"})
            attribs = {"d": path, "style": style}
            etree.SubElement(layer, "path", attribs)


def export_dimension(vals):
    # mandatory group codes : (10, 11, 13, 14, 20, 21, 23, 24) (x1..4, y1..4)
    # block_name: dimension definition for 10mm
    if vals.has_x1 and vals.has_x2 and vals.has_y1 and vals.has_y2:
        if vals.has_block_name:
            attribs = {
                inkex.addNS("href", "xlink"): "#%s" % (vals.block_name)
            }  # not use quote because name *D2 etc. changed to %2AD2
            tform = "translate(0 0)"
            # if vals.has_angle :
            #    tform += ' rotate(%f,%f,%f)' % (vals.angle,vals.x4,vals.y4)
            attribs.update({"transform": tform})
            etree.SubElement(layer, "use", attribs)
        else:
            # TODO: improve logic when INSERT in BLOCK
            dx = abs(vals.x1 - vals.x3)
            dy = abs(vals.y1 - vals.y3)
            if (vals.x1 == vals.x4) and dx > 0.00001:
                d = dx / scale
                dy = 0
                path = "M %f,%f %f,%f" % (vals.x1, vals.y1, vals.x3, vals.y1)
            elif (vals.y1 == vals.y4) and dy > 0.00001:
                d = dy / scale
                dx = 0
                path = "M %f,%f %f,%f" % (vals.x1, vals.y1, vals.x1, vals.y3)
            else:
                return
            attribs = {
                "d": path,
                "style": style
                + "; marker-start: url(#DistanceX); marker-end: url(#DistanceX); stroke-width: 0.25px",
            }
            etree.SubElement(layer, "path", attribs)
            x = vals.x2
            y = vals.y2
            size = 12  # default fontsize in px
            if vals.has_mtext:
                if vals.mtext in DIMTXT:
                    size = scale * textscale * DIMTXT[vals.mtext]
                    if size < 2:
                        size = 2
            attribs = {
                "x": "%f" % x,
                "y": "%f" % y,
                "style": "font-size: %.3fpx; fill: %s; font-family: %s; text-anchor: middle; text-align: center"
                % (size, color, options.font),
            }
            if dx == 0:
                attribs.update({"transform": "rotate (%f %f %f)" % (-90, x, y)})
            node = etree.SubElement(layer, "text", attribs)
            tspan = node.add(inkex.Tspan())
            tspan.set("sodipodi:role", "line")
            tspan.text = str(float("%.2f" % d))


def export_insert(vals):
    # mandatory group codes : (2, 10, 20) (block name, x, y)
    # TODO: repeat by row and column
    # (times,interval)= row(70,44), column(71,45)

    if vals.has_block_name and vals.has_x1 and vals.has_y1:
        # vals are through adjust_coords except block
        # block (x,y)=(0,0) : (scale*x-xmin, height-scale*y-ymin)
        # translate(move x units,move y units)
        # 2021.6  translate..ok  scale..x  rotate X
        # as scale, the line is wider ->same width  -> you should fix
        global height
        cx = scale * xmin  # transorm-origin:
        cy = scale * ymin + height  # center of rotation

        x = vals.x1 + scale * xmin
        y = vals.y1 - scale * ymin - height
        ixscale = iyscale = 1
        if vals.has_insert_scale_y:
            ixscale = vals.insert_scale_x
        if vals.has_insert_scale_y:
            iyscale = vals.insert_scale_y
        x += cx * (iyscale - 1)
        y -= cy * (iyscale - 1)

        elem = layer.add(inkex.Use())
        elem.set(
            inkex.addNS("href", "xlink"),
            "#" + quote(vals.block_name.replace(" ", "_").encode("utf-8")),
        )

        # add style stroke-width=1px for reducing thick line
        fwide = abs(0.5 / ixscale)  # better to use w/ixscale
        elem.style["stroke-width"] = "%.3fpx" % fwide

        elem.transform.add_translate(x, y)
        if vals.has_insert_scale_x and vals.has_insert_scale_y:
            elem.transform.add_scale(ixscale, iyscale)
        if vals.has_angle:
            rotated_angle = vals.angle
            if ixscale * iyscale > 0:
                rotated_angle = 360 - rotated_angle
            elem.transform.add_rotate(rotated_angle, -cx, cy)


def export_block(vals):
    # mandatory group codes : (2) (block name)
    if vals.has_block_name:
        global block
        block = etree.SubElement(
            defs, "symbol", {"id": vals.block_name.replace(" ", "_")}
        )


def export_endblk(vals):
    global block
    block = defs  # initiallize with dummy


def export_attdef(vals):
    # mandatory group codes : (1, 2) (default, tag)
    if vals.has_default and vals.has_tag:
        vals.text_list.append(vals.tag)
        export_mtext(vals)


def generate_ellipse(xc, yc, xm, ym, w, a1, a2):
    rm = math.sqrt(xm * xm + ym * ym)
    a = math.atan2(ym, xm)  # x-axis-rotation
    diff = (a2 - a1 + 2 * math.pi) % (2 * math.pi)
    if abs(diff) > 0.0000001 and abs(diff - 2 * math.pi) > 0.0000001:  # open arc
        large = 0  # large-arc-flag
        if diff > math.pi:
            large = 1
        xt = rm * math.cos(a1)
        yt = w * rm * math.sin(a1)
        x1 = xt * math.cos(a) - yt * math.sin(a)
        y1 = xt * math.sin(a) + yt * math.cos(a)
        xt = rm * math.cos(a2)
        yt = w * rm * math.sin(a2)
        x2 = xt * math.cos(a) - yt * math.sin(a)
        y2 = xt * math.sin(a) + yt * math.cos(a)
        path = "M %f,%f A %f,%f %f %d 0 %f,%f" % (
            xc + x1,
            yc - y1,
            rm,
            w * rm,
            -180.0 * a / math.pi,
            large,
            xc + x2,
            yc - y2,
        )
    else:  # closed arc
        path = "M %f,%f A %f,%f %f 0, 0 %f,%f A %f,%f %f 0, 0 %f,%f z" % (
            xc + xm,
            yc - ym,
            rm,
            w * rm,
            -180.0 * a / math.pi,
            xc - xm,
            yc + ym,
            rm,
            w * rm,
            -180.0 * a / math.pi,
            xc + xm,
            yc - ym,
        )
    attribs = {"d": path, "style": style}
    etree.SubElement(layer, "path", attribs)


def generate_gcodetools_point(xc, yc):
    elem = layer.add(inkex.PathElement())
    elem.style = "stroke:none;fill:#ff0000"
    elem.set("inkscape:dxfpoint", "1")
    elem.path = (
        "m %s,%s 2.9375,-6.34375 0.8125,1.90625 6.84375,-6.84375 0,0 0.6875,0.6875 -6.84375,6.84375 1.90625,0.8125 z"
        % (xc, yc)
    )


#   define DXF Entities and specify which Group Codes to monitor


class DxfInput(inkex.InputExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab", default="options")
        pars.add_argument("--scalemethod", default="manual")
        pars.add_argument("--scale", default="1.0")
        pars.add_argument("--textscale", default="1.0")
        pars.add_argument("--xmin", default="0.0")
        pars.add_argument("--ymin", default="0.0")
        pars.add_argument("--gcodetoolspoints", default=False, type=inkex.Boolean)
        pars.add_argument("--encoding", dest="input_encode", default="latin_1")
        pars.add_argument("--font", default="Arial")

    def load(self, stream):
        return stream

    def effect(self):
        global options
        global defs
        global entity
        global seqs
        global style
        global layer
        global scale
        global textscale
        global color
        global extrude
        global xmin
        global ymin
        global height
        global DIMTXT
        global block
        global svg
        global style_font3
        global style_direction
        global be_extrude

        options = self.options

        doc = self.get_template(width=210 * 96 / 25.4, height=297 * 96 / 25.4)
        svg = doc.getroot()
        defs = svg.defs
        marker = etree.SubElement(
            defs,
            "marker",
            {
                "id": "DistanceX",
                "orient": "auto",
                "refX": "0.0",
                "refY": "0.0",
                "style": "overflow:visible",
            },
        )
        etree.SubElement(
            marker,
            "path",
            {
                "d": "M 3,-3 L -3,3 M 0,-5 L  0,5",
                "style": "stroke:#000000; stroke-width:0.5",
            },
        )
        pattern = etree.SubElement(
            defs,
            "pattern",
            {
                "id": "Hatch",
                "patternUnits": "userSpaceOnUse",
                "width": "8",
                "height": "8",
                "x": "0",
                "y": "0",
            },
        )
        etree.SubElement(
            pattern,
            "path",
            {
                "d": "M8 4 l-4,4",
                "stroke": "#000000",
                "stroke-width": "0.25",
                "linecap": "square",
            },
        )
        etree.SubElement(
            pattern,
            "path",
            {
                "d": "M6 2 l-4,4",
                "stroke": "#000000",
                "stroke-width": "0.25",
                "linecap": "square",
            },
        )
        etree.SubElement(
            pattern,
            "path",
            {
                "d": "M4 0 l-4,4",
                "stroke": "#000000",
                "stroke-width": "0.25",
                "linecap": "square",
            },
        )

        def _get_line(strip=True):
            line = self.document.readline()
            if strip:
                line = line.strip()
            else:
                line = line.rstrip(b"\r\n")
            return line.decode(options.input_encode)

        def get_line():
            l1 = _get_line()
            if l1 in ("1", "2", "3", "6", "7", "8"):  # text value
                # text entity, do not strip
                return l1, _get_line(False)
            return l1, _get_line()

        def get_binary_line() -> Tuple[str, str]:
            stream = self.document
            # Data types inferred from
            # https://help.autodesk.com/view/OARX/2023/ENU/?guid=GUID-2553CF98-44F6-4828-82DD-FE3BC7448113

            try:
                key = struct.unpack("<h", stream.read(2))[0]
                if (
                    key <= 9
                    or 100 <= key <= 109
                    or 300 <= key <= 309
                    or 320 <= key <= 369
                    or 390 <= key <= 399
                    or 410 <= key <= 419
                    or 430 <= key <= 439
                    or 470 <= key <= 481
                    or 999 <= key <= 1009
                ):
                    string_bytes = bytearray()
                    while True:
                        byte = stream.read(1)
                        if byte == b"\x00" or len(byte) == 0:
                            value = string_bytes.decode().replace("\t", "")
                            break
                        string_bytes.extend(byte)
                elif (
                    10 <= key <= 59
                    or 110 <= key <= 149
                    or 210 <= key <= 239
                    or 460 <= key <= 469
                    or 1010 <= key <= 1059
                ):
                    value = struct.unpack("d", stream.read(8))[0]
                elif (
                    60 <= key <= 79
                    or 170 <= key <= 179
                    or 270 <= key <= 289
                    or 370 <= key <= 389
                    or 400 <= key <= 409
                    or 1060 <= key <= 1070
                ):
                    value = struct.unpack("<h", stream.read(2))[0]
                elif (
                    80 <= key <= 99
                    or 420 <= key <= 429
                    or 440 <= key <= 459
                    or key == 1071
                ):
                    value = struct.unpack("<i", stream.read(4))[0]
                elif 160 <= key <= 169:
                    value = struct.unpack("<q", stream.read(8))[0]
                elif 290 <= key <= 299:
                    value = struct.unpack("?", stream.read(1))[0]
                elif 310 <= key <= 319:
                    # binary data
                    length = struct.unpack("<B", stream.read(1))[0]
                    value = stream.read(length)
                else:
                    inkex.errormsg(
                        _(
                            "Encountered key of unknown format, id={0} while parsing DXF. This is probably a bug."
                        ).format(key)
                    )
                return str(key), str(value)
            except struct.error:
                return "", ""
                # EOF

        def get_group(group):
            line = get_line()
            if line[0] == group:
                return float(line[1])
            return 0.0

        xmax = xmin = ymin = 0.0
        ltscale = 1.0  # $LTSCALE:global scale of line-style
        height = 297.0 * 96.0 / 25.4  # default A4 height in pixels
        measurement = 0  # default inches
        flag = 0  # (0, 1, 2, 3, 4) = (none, LAYER, LTYPE, DIMTXT, STYLE)
        layer_colors = {}  # store colors by layer
        layer_nodes = {}  # store nodes by layer
        linetypes = {}  # store linetypes by name
        DIMTXT = {}  # store DIMENSION text sizes
        # style_name = {}     # style name
        style_font3 = {}  # style font 1byte
        style_font4 = {}  # style font 2byte
        style_direction = {}  # style display direction
        be_extrude = False
        first_line = _get_line()

        if first_line == "AutoCAD Binary DXF":
            magic = self.document.read(2)
            if not magic == b"\x1a\x00":
                inkex.AbortExtension(_("Binary DXF file has unexpected header."))
            get_line = get_binary_line
            line = [first_line, ""]
        else:
            line = [first_line, _get_line()]

        inENTITIES = False
        style_name = "*"
        layername = None
        linename = None
        stylename = None
        style_name = None
        errno = 0
        pdmode_err = False
        while line[0] and line[1] != "BLOCKS":
            line = get_line()
            if line[1] == "ENTITIES":  # no BLOCK SECTION
                inENTITIES = True
                break
            if line[1] == "$PDMODE" and not pdmode_err:
                inkex.errormsg("$PDMODE is ignored. A point is displayed as normal.")
                pdmode_err = True
            if options.scalemethod == "file":
                if line[1] == "$MEASUREMENT":
                    measurement = get_group("70")
            elif options.scalemethod == "auto":
                if line[1] == "$EXTMIN":
                    xmin = get_group("10")
                    ymin = get_group("20")
                if line[1] == "$EXTMAX":
                    xmax = get_group("10")
            if line[1] == "$LTSCALE":
                ltscale = get_group("40")
            if flag == 1 and line[0] == "2":
                layername = line[1]
                layer_nodes[layername] = svg.add(inkex.Layer.new(layername))
            if flag == 2 and line[0] == "2":
                linename = line[1]
                linetypes[linename] = []
            if flag == 3 and line[0] == "2":
                stylename = line[1]
            if flag == 4 and line[0] == "2":
                style_name = line[1]
                style_font3[style_name] = []
                style_font4[style_name] = []
                style_direction[style_name] = []
            if line[0] == "2" and line[1] == "LAYER":
                flag = 1
            if line[0] == "2" and line[1] == "LTYPE":
                flag = 2
            if line[0] == "2" and line[1] == "DIMSTYLE":
                flag = 3
            if line[0] == "2" and line[1] == "STYLE":
                flag = 4
            if flag == 1 and line[0] == "62":
                if layername is None:
                    errno = 1
                    break
                layer_colors[layername] = int(line[1])
            if flag == 2 and line[0] == "49":
                if linename is None:
                    errno = 2
                    break
                linetypes[linename].append(float(line[1]))
            if flag == 3 and line[0] == "140":
                if stylename is None:
                    errno = 3
                    break
                DIMTXT[stylename] = float(line[1])
            if flag == 4 and line[0] == "3":
                if style_name is None:
                    errno = 4
                    break
                style_font3[style_name].append(line[1])
            if flag == 4 and line[0] == "4":
                if style_name is None:
                    errno = 4
                    break
                style_font4[style_name].append(line[1])
            if flag == 4 and line[0] == "70":  # not no of STYLE
                if style_name != "*":
                    style_direction[style_name] = int(line[1])
            if line[0] == "0" and line[1] == "ENDTAB":
                flag = 0
        if errno != 0:
            if errno == 1:
                errMsg = "LAYER"
            elif errno == 2:
                errMsg = "LTYPE"
            elif errno == 3:
                errMsg = "DIMSTYLE"
            else:  # errno == 4
                errMsg = "STYLE"
            inkex.errormsg(
                "Import stopped. DXF is incorrect.\ngroup code 2 ("
                + errMsg
                + ") is missing"
            )
            self.document = doc
            return

        if options.scalemethod == "file":
            scale = 25.4  # default inches
            if measurement == 1.0:
                scale = 1.0  # use mm
        elif options.scalemethod == "auto":
            scale = 1.0
            if xmax > xmin:
                scale = 210.0 / (xmax - xmin)  # scale to A4 width
        else:
            scale = float(options.scale)  # manual scale factor
            xmin = float(options.xmin)
            ymin = float(options.ymin)
        svg.desc = "%s - scale = %f, origin = (%f, %f), method = %s" % (
            os.path.basename(options.input_file),
            scale,
            xmin,
            ymin,
            options.scalemethod,
        )
        scale *= 96.0 / 25.4  # convert from mm to pixels
        textscale = float(options.textscale)

        if "0" not in layer_nodes:
            layer_nodes["0"] = svg.add(inkex.Layer.new("0"))

            layer_colors["0"] = 7

        for linename in linetypes.keys():  # scale the dashed lines
            linetype = ""
            for length in linetypes[linename]:
                if length == 0:  # test for dot
                    linetype += " 0.5,"
                else:
                    linetype += "%.4f," % math.fabs(length * scale * ltscale)
            if linetype == "":
                linetypes[linename] = "stroke-linecap: round"
            else:
                linetypes[linename] = "stroke-dasharray:" + linetype

        entity = ""
        block = defs  # initiallize with dummy
        while line[0] and (line[1] != "ENDSEC" or not inENTITIES):
            line = get_line()
            if line[1] == "ENTITIES":
                inENTITIES = True
            if entity and vals.is_valid(line[0]):
                seqs.append(line[0])  # list of group codes
                if line[0] in ("1", "2", "3", "6", "7", "8"):  # text value
                    # TODO: if add funs of export_mtext, delete the line
                    val = line[1].replace(r"\~", " ")
                    # val = re.sub(r"\\A.*;", "", val)
                    # val = re.sub(r'\\H.*;', '', val)
                    val = re.sub(r"\^I", "", val)
                    val = re.sub(r"{\\L", "", val)
                    # val = re.sub(r'}', '', val)  {\\C; '}' in mtext
                    val = re.sub(r"\\S.*;", "", val)
                    val = re.sub(r"\\W.*;", "", val)
                    val = val
                    val = re.sub(r"\\U\+([0-9A-Fa-f]{4})", re_hex2unichar, val)
                elif line[0] in ("62", "70", "92", "93"):
                    val = int(line[1])
                else:  # unscaled float value
                    val = float(line[1])
                vals[line[0]].append(val)
            elif has_export(line[1]):
                if has_export(entity):
                    if block != defs:  # in a BLOCK
                        layer = block
                    elif vals.has_layer_name:  # use Common Layer Name
                        if not vals.layer_name:
                            vals.layer_name = "0"  # use default name
                        if vals.layer_name not in layer_nodes:
                            # attribs = {inkex.addNS('groupmode','inkscape') :
                            #    'layer', inkex.addNS('label','inkscape') : '%s' % vals.layer_name}
                            # layer_nodes[vals.layer_name] = etree.SubElement(doc.getroot(), 'g', attribs)
                            layer_nodes[vals.layer_name] = svg.add(
                                inkex.Layer.new(vals.layer_name)
                            )
                        layer = layer_nodes[vals.layer_name]
                    color = "#000000"  # default color
                    if vals.has_layer_name:
                        if vals.layer_name in layer_colors:
                            color = get_rgbcolor(layer_colors[vals.layer_name], color)
                    if vals.has_color:  # Common Color Number
                        color = get_rgbcolor(vals.color, color)
                    style = formatStyle({"stroke": "%s" % color, "fill": "none"})
                    w = 0.5  # default lineweight for POINT
                    if vals.has_line_weight:  # Common Lineweight
                        if vals.line_weight > 0:
                            w = 96.0 / 25.4 * vals.line_weight / 100.0
                            w *= scale  # real wide : line_weight /144 inch
                            if w < 0.5:
                                w = 0.5
                            if (
                                block == defs
                            ):  # not in a BLOCK for INSERT except stroke-width 2021.july
                                style = formatStyle(
                                    {
                                        "stroke": "%s" % color,
                                        "fill": "none",
                                        "stroke-width": "%.3f" % w,
                                    }
                                )
                    if vals.has_line_type:  # Common Linetype
                        if vals.line_type in linetypes:
                            style += ";" + linetypes[vals.line_type]
                    extrude = 1.0
                    if vals.has_extrude:
                        if (entity != "LINE") and (entity != "POINT"):
                            extrude = float(vals.extrude)
                            if extrude < 1.0:
                                be_extrude = True

                    vals.adjust_coords(xmin, ymin, scale, extrude, height)

                    if extrude == -1.0:  # reflect angles
                        if vals.has_angle and vals.has_angle2:
                            vals.angle2, vals.angle = (
                                180.0 - vals.angle,
                                180.0 - vals.angle2,
                            )
                    exporter = get_export(entity)
                    if exporter:
                        if entity == "POINT":
                            exporter(vals, w)
                        else:
                            exporter(vals)

                if line[1] == "POLYLINE":
                    inVertexs = False
                    entity = "LWPOLYLINE"
                    vals = ValueConstruct()
                    seqs = []
                    flag70 = 0  # default closed-line or not
                    val8 = "0"  # default layer name
                    val10 = 0  # x
                    val20 = 0  # y
                    val42 = 0  # bulge
                    valid = True
                    while line[0] and (line[1] != "SEQEND"):
                        line = get_line()
                        if line[1] == "VERTEX":
                            if inVertexs == True:
                                if valid:
                                    seqs.append("10")
                                    vals["10"].append(val10)
                                    seqs.append("20")
                                    vals["20"].append(val20)
                                    seqs.append("42")
                                    vals["42"].append(val42)
                                    val42 = 0
                            inVertexs = True
                            valid = True
                        if inVertexs == False:
                            if line[0] == "6":  # 6:line style
                                seqs.append(line[0])
                                vals[line[0]].append(line[1])
                            if line[0] == "8":  # 8:layer
                                val8 = line[1]
                            if line[0] == "62":  # color
                                vals[line[0]].append(int(line[1]))
                            if line[0] == "70":  # flag
                                flag70 = int(line[1])
                        else:
                            if line[0] == "70":
                                if int(line[1]) == 16:  # control point
                                    valid = False
                            if line[0] in ("10", "20", "42"):  # vertexs
                                val = float(line[1])
                                if line[0] == "10":
                                    val10 = val
                                elif line[0] == "20":
                                    val20 = val
                                else:
                                    val42 = val
                    if valid:
                        seqs.append("8")  # layer_name
                        vals["8"].append(val8)
                        seqs.append("10")
                        vals["10"].append(val10)
                        seqs.append("20")
                        vals["20"].append(val20)
                        seqs.append("42")  # bulge
                        vals["42"].append(val42)
                    seqs.append("70")  # closed line?
                    vals["70"].append(flag70)
                    continue

                entity = line[1]
                vals = ValueConstruct()
                seqs = []

        #     for debug
        # tree = etree.ElementTree(svg)
        # tree.write('c:\Python\svgCH2.xml')
        if be_extrude:
            inkex.errormsg(
                _(
                    "An object that has the extrude parameter set was detected. "
                    "The imported figure may be displayed incorrectly."
                )
            )
        self.document = doc


def get_export(opt):
    return globals().get("export_" + opt.lower(), None)


def has_export(opt):
    return get_export(opt) is not None


if __name__ == "__main__":
    DxfInput().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2009 Karlisson Bezerra, contato@nerdson.com
#               2021 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Splits a text element into lines, words, chars.
Supports all text elements that Inkscape can create, such as normal text, shape-inside (SVG2),
flowroot (SVG1.2), inline-size, manual kerns, and nested tspans (with possibly different kerns)

The code is structured as followed. For each selected text element:
 - preprocess_text_element duplicates the element, converts flowroots to plain text elements,
   and simplifies manual kerning if requested (only for split_words and split_chars, for all,
   possibly nested, children) using simplify_nested_tspans.
 - if split lines: split_lines copies all top-level tspans from the previous step into their own
   text element, which is otherwise a duplicate of the original text element (thus preserving
   style and transforms), see append_splitted_element
 - if split words or chars: split_words_or_chars: the text is recursively processed. For each tspan,
   the content and tail is split (words: at spaces, chars: after each character) into their own
   tspan, again using append_splitted_element. The method keeps track of the horizontal and vertical
   coordinate, incrementing it with the number of characters and a multiple of font size.
"""

import re as regex
from typing import Union, Callable

import inkex
from inkex import TextElement, FlowRoot, FlowPara, Tspan, Rectangle, ShapeElement
from inkex.units import parse_unit
from inkex.localization import inkex_gettext as _

TextLike = Union[FlowRoot, TextElement]


class TextSplit(inkex.EffectExtension):
    """Split text up."""

    def __init__(self):
        """Initialize State machine"""
        super().__init__()
        self.mode: Callable
        self.separation: float = 1
        self.fs_multiplier: float = 0.25
        self.current_x: float = 0
        self.current_y: float = 0
        self.process_kerns: bool = True
        self.current_root: TextLike
        self.current_fontsize: float = 0

    def add_arguments(self, pars):
        pars.add_argument("--tab", help="The selected UI tab when OK was pressed")
        pars.add_argument(
            "-t",
            "--splittype",
            default="line",
            choices=["letter", "word", "line"],
            help="type of split",
        )
        pars.add_argument(
            "-p",
            "--preserve",
            type=inkex.Boolean,
            default=True,
            help="Preserve original",
        )
        pars.add_argument(
            "-s",
            "--separation",
            type=float,
            default=1,
            help="Threshold for separating text with manual kerns in multiples of"
            "font-size",
        )

    def effect(self):
        """Applies the effect"""

        split_type = self.options.splittype
        preserve = self.options.preserve

        # checks if the selected elements are text nodes
        for elem in self.svg.selection.filter_nonzero(TextElement, FlowRoot):
            try:
                self.separation = self.options.separation
                if split_type == "line":
                    node = self.split_lines(elem)
                elif split_type == "word":
                    self.mode = self.process_plain_words
                    node = self.split_words_or_chars(elem)
                else:
                    self.separation = 0
                    self.mode = self.process_plain_chars
                    node = self.split_words_or_chars(elem)

                node.getparent().remove(node)

                if not preserve and node is not None:
                    elem.getparent().remove(elem)
            except TypeError as err:
                inkex.errormsg(err)  # if an element can not be processed

    @staticmethod
    def get_font_size(element):
        """get the font size of an element"""
        return element.get_computed_style("font-size")

    @staticmethod
    def get_line_height(element: ShapeElement):
        """get the line height of an element"""
        return element.get_line_height_uu()

    def simplify_child_tspans(self, element: TextElement):
        """Checks all child tspans if they have manual kerns.
        If it does, try to find words (characters with a distance > separation * font-size).
        Then concatenate the words with spaces, set this string as a new text and"""
        for child in list(element):
            # process manual kerns
            if not isinstance(child, Tspan):
                continue
            xvals = list(
                map(float, filter(len, regex.split(r"[,\s]", child.get("x") or "")))
            )
            content = child.text
            if content not in [None, ""] and len(xvals) >= 2:
                fsize = self.get_font_size(child)
                separation = self.separation * fsize
                current_word_start = 0
                for i in range(1, max(len(content), len(xvals))):
                    if i >= len(content) - 1 or i >= len(xvals) - 1:
                        # consume the entire remaining string
                        i = len(content)
                    if i == len(content) or abs(xvals[i] - xvals[i - 1]) > separation:
                        wordspan = Tspan(x=str(xvals[current_word_start]))
                        wordspan.text = content[current_word_start:i]
                        child.add(wordspan)
                        current_word_start = i
                child.pop("x")
                child.text = None
            # process child elements
            self.simplify_child_tspans(child)

    def preprocess_text_element(self, element: TextElement):
        """Processes a text element and returns an element containing tspans with x and y coordinate,
        possibly nested (for Inkscape-type kerning), so that the actual splitting can work as if the
        text was a simple text. Manual kerns (one x value per letter) are converted to spaces
        if requested (not necessary for "split characters")"""

        oldelement = element
        if isinstance(element, FlowRoot):
            element = TextElement()
            oldelement.addnext(element)
            element.style = oldelement.style
            element.transform = oldelement.transform
            flowref = oldelement.findone("svg:flowRegion")[0]
            if isinstance(flowref, Rectangle):
                flowx = element.unittouu(flowref.get("x"))
                flowy = element.unittouu(float(flowref.get("y")))
                first = True
            else:
                raise TypeError(
                    _(
                        "Element {} uses a flow region that is not a rectangle. "
                        "First unflow text."
                    ).format(element.get_id())
                )
            for child in oldelement:
                if isinstance(child, FlowPara):
                    # convert the flowpara "line" (note: no automatic wrapping)
                    # to a tspan and set the y coordinate.
                    # future FlowRoot improvements could add a better conversion.
                    newchild = Tspan()
                    element.append(newchild)
                    newchild.text = child.text
                    newchild.style = child.style
                    newchild.transform = child.transform
                    newchild.set("x", flowx)
                    if first:
                        flowy += self.get_font_size(child) * 1.25
                        first = False
                    else:
                        flowy += self.get_line_height(child)
                    newchild.set("y", str(flowy))

        else:
            element = oldelement.duplicate()
            oldelement.getparent().append(element)

        element.style.pop("shape-inside", None)

        # Real support for RTL text is missing, but we can emulate it by just removing the
        # attribute. However, line breaks will be misaligned.
        element.style.pop("direction", None)
        for child in element:
            child.style.pop("direction", None)

        if self.process_kerns:
            self.simplify_child_tspans(element)
        return element

    def append_splitted_element(self, text, prototype=None):
        """Creates a new text element, sibling to self.current_root, at (self.current_x,
        self.current_y) with content text.

        text: either a Tspan that should be moved to a new text element - in this case, text is
            a direct child of element; or a string
        prototype: if text is a string, style and transform will be taken from prototype
        """

        if isinstance(text, Tspan) and text.getparent() == self.current_root:
            # we just move the tspan to a new text element.
            elem = self.current_root.duplicate()
            elem.remove_all(Tspan)
            elem.append(text)
            elem.set("x", text.get("x"))
            elem.set("y", text.get("y"))
        else:
            elem = TextElement(x=str(self.current_x), y=str(self.current_y))
            # transfer the style from all parents, including the text element (if there's a style to
            # the text element's parent applied, it will be duplicated, but that doesn't really
            # matter)
            elem.style = prototype.specified_style()
            # the element will be appended to the parent of element, but there might be nested
            # tspans between the prototype and the element. The next line says
            # "compose transforms until you reach the parent of element"
            elem.transform = (
                -self.current_root.getparent().transform
            ) @ prototype.composed_transform()
            tsp = Tspan(x=str(self.current_x), y=str(self.current_y))
            tsp.text = text
            elem.add(tsp)
        self.current_root.addnext(elem)

    def split_lines(self, element: TextLike) -> TextElement:
        """Splits a text into its lines"""
        self.process_kerns = False
        preprocessed = self.preprocess_text_element(element)
        self.current_root = preprocessed
        # Now we only have to copy each tspan into its own text element.
        for child in list(preprocessed):
            self.append_splitted_element(child)

        return preprocessed

    def process_plain_text(self, element, splitted):
        """Appends new text elements to as sibling root for each element of splitted, starting at
        self.current_x, self.current_y, incrementing those, with prototype element (that
        styles and transforms will be taken from)"""
        if splitted is None:
            return
        for word in splitted:
            if word != "":
                self.append_splitted_element(word, element)
            # +1 since for words, we lost a space
            self.current_x += (
                self.current_fontsize * (len(word) + 1) * self.fs_multiplier
            )

    def process_plain_words(self, element, text):
        """Calls process_plain_text for splitting words"""
        self.fs_multiplier = 0.4
        if text is not None:
            self.process_plain_text(element, text.split(" "))

    def process_plain_chars(self, element, text):
        """Calls process_plain_text for splitting characters"""
        self.fs_multiplier = 0.25
        self.process_plain_text(element, text)

    def split_words_or_chars(self, element: TextLike) -> TextElement:
        """Splits a text into its lines"""
        self.process_kerns = True
        preprocessed = self.preprocess_text_element(element)

        def process_element(element) -> float:
            elem_coords = {
                i: (
                    element.root.unittouu(element.get(i))
                    if element.get(i) is not None
                    else None
                )
                for i in "xy"
            }
            if elem_coords["x"] is not None:
                self.current_x = elem_coords["x"]
            if elem_coords["y"] is not None:
                self.current_y = elem_coords["y"]
            self.current_fontsize = self.get_font_size(element)
            current_x = self.mode(element, element.text)

            for elem in element:
                if isinstance(elem, Tspan):
                    current_x = process_element(elem)
                current_x = self.mode(element, elem.tail)
            return current_x

        self.current_root = preprocessed
        process_element(preprocessed)
        return preprocessed


if __name__ == "__main__":
    TextSplit().run()



#!/usr/bin/env python3
"""
Optimise PNG file using optipng
"""

import os
import inkex
from inkex.extensions import TempDirMixin
from inkex.command import ProgramRunError, call


class PngOutput(TempDirMixin, inkex.RasterOutputExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        # Lossless options
        pars.add_argument("--interlace", type=inkex.Boolean, default=False)
        pars.add_argument("--level", type=int, default=0)
        # Lossy options
        pars.add_argument("--bitdepth", type=inkex.Boolean, default=False)
        pars.add_argument("--color", type=inkex.Boolean, default=False)
        pars.add_argument("--palette", type=inkex.Boolean, default=False)

    def load(self, stream):
        """Load the PNG file (prepare it for optipng)"""
        self.png_file = os.path.join(self.tempdir, "input.png")
        with open(self.png_file, "wb") as fhl:
            fhl.write(stream.read())

    def save(self, stream):
        """Pass the PNG file to optipng with the options"""
        options = {
            "o": self.options.level,
            "i": int(self.options.interlace),
            "nb": not self.options.bitdepth,
            "nc": not self.options.color,
            "np": not self.options.palette,
        }
        try:
            call("optipng", self.png_file, oldie=True, clobber=True, **options)
        except ProgramRunError as err:
            if "IDAT recoding is necessary" in err.stderr.decode("utf-8"):
                raise inkex.AbortExtension(
                    _(
                        "The optipng command failed, possibly due to a mismatch of the"
                        "interlacing and compression level options. Please try to disable "
                        '"Interlaced" or set "Level" to 1 or higher.'
                    )
                )
            else:
                raise inkex.AbortExtension(
                    _("The optipng command failed with the following message:")
                    + "\n"
                    + err.stderr.decode("utf-8")
                )

        if os.path.isfile(self.png_file):
            with open(self.png_file, "rb") as fhl:
                stream.write(fhl.read())


if __name__ == "__main__":
    PngOutput().run()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Maren Hachmann <moini>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=arguments-differ
"""
Provide extra utility to each svg element type specific to its type.

This is useful for having a common interface for each element which can
give path, transform, and property access easily.
"""

from __future__ import annotations
import math

from typing import List, Optional

from lxml import etree

from inkex.deprecated.meta import deprecate

from ..styles import StyleSheet
from ..transforms import BoundingBox, Vector2d, VectorLike, DirectedLineSegment

from ._base import BaseElement


class Defs(BaseElement):
    """A header defs element, one per document"""

    tag_name = "defs"


class StyleElement(BaseElement):
    """A CSS style element containing multiple style definitions"""

    tag_name = "style"

    def set_text(self, content):
        """Sets the style content text as a CDATA section"""
        self.text = etree.CDATA(str(content))

    def stylesheet(self):
        """Return the StyleSheet() object for the style tag"""
        return StyleSheet(self.text, callback=self.set_text)


class Script(BaseElement):
    """A javascript tag in SVG"""

    tag_name = "script"

    def set_text(self, content):
        """Sets the style content text as a CDATA section"""
        self.text = etree.CDATA(str(content))


class Desc(BaseElement):
    """Description element"""

    tag_name = "desc"


class Title(BaseElement):
    """Title element"""

    tag_name = "title"


class NamedView(BaseElement):
    """The NamedView element is Inkscape specific metadata about the file"""

    tag_name = "sodipodi:namedview"

    current_layer = property(lambda self: self.get("inkscape:current-layer"))

    @property
    def center(self):
        """Returns view_center in terms of document units"""
        return Vector2d(
            self.root.viewport_to_unit(self.get("inkscape:cx") or 0),
            self.root.viewport_to_unit(self.get("inkscape:cy") or 0),
        )

    def get_guides(self):
        """Returns a list of guides"""
        return self.findall("sodipodi:guide")

    def add_guide(self, position, orient=True, name=None) -> Guide:
        """Creates a new guide in this namedview

        .. versionadded:: 1.3

        Args:
            position: a float containing the y position for ``orient is True``, or
                the x position for ``orient is False``. The position is specified in the
                post-1.0 coordinate system, i.e. y=0 is at the top left of the viewbox,
                positive y axis pointing down.

                Alternatively, the position may be given as Tuple (or VectorLike)
            orient: True for horizontal, False for Vertical

                alternatively: Tuple / Vector specifying x and y coordinates of the
                normal vector of the guide, or the (clockwise) angle between the
                horizontal axis and the guide. Defaults to True (horizontal)
            name: label of the guide

        Returns:
            the created guide
        """
        elem = self.add(Guide())

        if orient is True:
            elem.set_position(0, position, (0, -1))
        elif orient is False:
            elem.set_position(position, self.root.viewbox_height, (1, 0))
        else:
            pos = Vector2d(position)
            elem.set_position(pos.x, pos.y, orient)
        if name:
            elem.set("inkscape:label", str(name))
        return elem

    @deprecate
    def new_guide(self, position, orient=True, name=None):
        """
        .. deprecated:: 1.3
            Use :func:`add_guide` instead.

        Creates a new guide in this namedview

        Args:
            position: a float containing the y position for ``orient is True``, or
                the x position for ``orient is False``

                .. versionchanged:: 1.2
                    Alternatively, the position may be given as Tuple (or VectorLike)
            orient: True for horizontal, False for Vertical

                .. versionchanged:: 1.2
                    Tuple / Vector specifying x and y coordinates of the normal vector
                    of the guide.
            name: label of the guide

        Returns:
            the created guide
        """
        if orient is True:
            elem = Guide().move_to(0, position, (0, 1))
        elif orient is False:
            elem = Guide().move_to(position, 0, (1, 0))
        else:
            elem = Guide().move_to(*position, orient)
        if name:
            elem.set("inkscape:label", str(name))
        return self.add(elem)

    @deprecate
    def new_unique_guide(
        self, position: Vector2d, orientation: Vector2d
    ) -> Optional[Guide]:
        """
        .. deprecated:: 1.3
            Use :func:`add_unique_guide` instead.

        Add a guide iif there is no guide that looks the same.

        .. versionadded:: 1.2

        """
        elem = Guide().move_to(position[0], position[1], orientation)
        return self.add(elem) if self.get_similar_guide(elem) is None else None

    def add_unique_guide(
        self, position: Vector2d, orientation: Vector2d
    ) -> Optional[Guide]:
        """Add a guide iif there is no guide that looks the same.

        .. versionadded:: 1.3

        Args:
            position: Position as Tuple / Vector
            orientation: Tuple / Vector specifying x and y coordinates of the normal
                vector of the guide.
            name: label of the guide
        """
        elem = self.add(Guide()).set_position(position[0], position[1], orientation)
        if self.get_similar_guide(elem) is not None:
            self.remove(elem)
            return None
        return elem

    def get_similar_guide(self, other: Guide) -> Optional[Guide]:
        """Check if the namedview contains a guide that looks identical to one
        defined by (position, orientation) and is not identity (same element) as the
        first one. If such a guide exists, return it; otherwise, return None.

        .. versionadded:: 1.2
        """
        for guide in self.get_guides():
            if Guide.guides_coincident(guide, other) and guide != other:
                return guide
        return None

    def _get_pages(self) -> List[Page]:
        """Returns all page elements"""
        return self.findall("inkscape:page")

    def _equivalent_page(self) -> Page:
        """Returns an unrooted page based on the viewbox dimensions"""
        return Page.new(self.root.viewbox_width, self.root.viewbox_height, 0, 0)

    def get_pages(self) -> List[Page]:
        """Returns a list of pages within the document. For single page documents,
        a detached page element with dimensions according to the viewbox will be
        returned.

        .. versionadded:: 1.2

        .. versionchanged:: 1.3
            For single-page documents, this function now returns the viewbox
            dimensions.
        """
        pages = self._get_pages()
        if len(pages) < 2:
            return [self._equivalent_page()]
        return pages

    def new_page(self, x, y, width, height, label=None):
        """Creates a new page in this namedview. Always add pages through this
        function to ensure that single-page documents are treated correctly.

        .. versionadded:: 1.2

        .. versionchanged:: 1.3
            If none exists, a page element with the viewbox dimensions will be
            inserted before the new page.
        """
        if len(self._get_pages()) == 0:
            self.add(self._equivalent_page())
        elem = Page(width=width, height=height, x=x, y=y)
        if label:
            elem.set("inkscape:label", str(label))
        return self.add(elem)


class Guide(BaseElement):
    """An inkscape guide"""

    tag_name = "sodipodi:guide"

    @property
    def orientation(self) -> Vector2d:
        """Vector normal to the guide, in the pre-1.0 coordinate system (y axis upwards)

        .. versionadded:: 1.2
        """
        return Vector2d(self.get("orientation"), fallback=(1, 0))

    @property
    def angle(self) -> float:
        """(Clockwise) angle between the guide and the horizontal axis in degrees
        (i.e. what Inkscape 1.2+ shows as "Angle" in the guide properties)

        .. versionadded:: 1.3
        """
        return math.degrees(math.atan2(*self.orientation))

    is_horizontal = property(
        lambda self: self.orientation[0] == 0 and self.orientation[1] != 0
    )
    is_vertical = property(
        lambda self: self.orientation[0] != 0 and self.orientation[1] == 0
    )

    @property
    def raw_position(self) -> Vector2d:
        """Position of the guide handle.

        The y coordinate is flipped and relative to the bottom of the
        viewbox, this is a remnant of the pre-1.0 coordinate system
        """
        return Vector2d(self.get("position"), fallback=(0, 0))

    def point(self):
        """Use raw_position or position instead"""
        return self.raw_position

    point = property(deprecate(point))  # type: ignore

    @property
    def position(self) -> Vector2d:
        """Position of the guide handle in normal coordinates.

        This means (0,0) is at the top left corner of the viewbox,
        positive y axis pointing downwards.

        This function can only be used for guides which are attached to a root
        svg element.
        """
        pos = self.raw_position
        return Vector2d(pos.x, self.root.viewbox_height - pos.y)

    @classmethod
    def new(cls, pos_x, pos_y, angle, **attrs):
        guide = super().new(**attrs)
        guide.set_position(pos_x, pos_y, angle=angle)
        return guide

    def set_position(self, pos_x, pos_y, angle=None):
        """
        Move this guide to the given x,y position and optionally set its orientation.

        The coordinate system used is the post-1.0 coordinate system (origin in the
        top left corner, y axis pointing down), which also defines the sense of
        rotation.

        The guide must be rooted for this function to be used. Preferably, use
        :func:`inkex.elements._meta.add_guide` to create a new guide.

        .. versionadded:: 1.3

        Args:
            pos_x (Union[str, int, float]): x position of the guide's reference point
            pos_y (Union[str, int, float]): y position of the guide's reference point
            angle (Union[str, float, int, tuple, list], optional): Angle may be a
                string, float or integer, which will set the clockwise angle between the
                horizontal axis and the guide.

                Alternatively, it may be a pair of numbers (tuple) which will be set
                as normal vector.

                If not given at all, the orientation remains unchanged.
                Defaults to None.

        Returns:
            Guide: the modified guide
        """
        pos_y = self.root.viewbox_height - float(pos_y)
        self.set("position", f"{float(pos_x):g},{float(pos_y):g}")
        if isinstance(angle, str):
            if "," not in angle:
                angle = float(angle)

        if isinstance(angle, (float, int)):
            # Generate orientation from angle
            angle = (math.sin(math.radians(angle)), -math.cos(math.radians(angle)))

        if isinstance(angle, (tuple, list)) and len(angle) == 2:
            angle = ",".join(f"{i:g}" for i in [angle[0], -angle[1]])

        if angle is not None:
            self.set("orientation", angle)
        return self

    @deprecate
    def move_to(self, pos_x, pos_y, angle=None):
        """
        .. deprecated:: 1.3
            Use :func:`set_position` instead.

        Move this guide to the given x,y position,

        Angle may be a float or integer, which will change the orientation. Alternately,
        it may be a pair of numbers (tuple) which will set the orientation directly.
        If not given at all, the orientation remains unchanged.
        """
        self.set("position", f"{float(pos_x):g},{float(pos_y):g}")
        if isinstance(angle, str):
            if "," not in angle:
                angle = float(angle)

        if isinstance(angle, (float, int)):
            # Generate orientation from angle
            angle = (math.sin(math.radians(angle)), -math.cos(math.radians(angle)))

        if isinstance(angle, (tuple, list)) and len(angle) == 2:
            angle = ",".join(f"{i:g}" for i in angle)

        if angle is not None:
            self.set("orientation", angle)
        return self

    @staticmethod
    def guides_coincident(guide1, guide2):
        """Check if two guides defined by (position, orientation) and (opos, oor) look
        identical.

        Returns:
            True if the position lies on the other guide AND the guide is
            (anti)parallel to the other guide.

        .. versionadded:: 1.2
        """
        # normalize orientations first
        orientation = guide1.orientation / guide1.orientation.length
        oor = guide2.orientation / guide2.orientation.length

        position = guide1.raw_position
        opos = guide2.raw_position

        return (
            DirectedLineSegment(
                position, position + Vector2d(orientation[1], -orientation[0])
            ).perp_distance(*opos)
            < 1e-6
            and abs(abs(orientation[1] * oor[0]) - abs(orientation[0] * oor[1])) < 1e-6
        )


class Metadata(BaseElement):
    """Resource Description Framework (RDF) metadata"""

    tag_name = "metadata"

    doc_title = property(lambda self: self._first_text("dc:title"))
    description = property(lambda self: self._first_text("dc:description"))

    rights = property(lambda self: self._first_text("dc:rights/cc:Agent/dc:title"))
    creator = property(lambda self: self._first_text("dc:creator/cc:Agent/dc:title"))
    publisher = property(
        lambda self: self._first_text("dc:publisher/cc:Agent/dc:title")
    )
    contributor = property(
        lambda self: self._first_text("dc:contributor/cc:Agent/dc:title")
    )

    date = property(lambda self: self._first_text("dc:date"))
    source = property(lambda self: self._first_text("dc:source"))
    language = property(lambda self: self._first_text("dc:language"))
    relation = property(lambda self: self._first_text("dc:relation"))
    coverage = property(lambda self: self._first_text("dc:coverage"))
    identifier = property(lambda self: self._first_text("dc:identifier"))

    def _first_text(self, loc):
        """Get the work title"""
        elem = self.findone(f"rdf:RDF/cc:Work/{loc}")
        if elem:
            return elem.text
        return None

    @property
    def tags(self):
        return [
            elem.text
            for elem in self.findall("rdf:RDF/cc:Work/dc:subject/rdf:Bag/rdf:li")
        ]


class ForeignObject(BaseElement):
    """SVG foreignObject element"""

    tag_name = "foreignObject"


class Switch(BaseElement):
    """A switch element"""

    tag_name = "switch"


class Grid(BaseElement):
    """A namedview grid child"""

    tag_name = "inkscape:grid"


class Page(BaseElement):
    """A namedview page child

    .. versionadded:: 1.2
    """

    tag_name = "inkscape:page"

    width = property(lambda self: self.to_dimensionless(self.get("width") or 0))
    height = property(lambda self: self.to_dimensionless(self.get("height") or 0))
    x = property(lambda self: self.to_dimensionless(self.get("x") or 0))
    y = property(lambda self: self.to_dimensionless(self.get("y") or 0))

    @classmethod
    def new(cls, width, height, x, y):
        """Creates a new page element in the namedview"""
        page = super().new()
        page.move_to(x, y)
        page.set("width", width)
        page.set("height", height)
        return page

    def move_to(self, x, y):
        """Move this page to the given x,y position"""
        self.set("x", f"{float(x):g}")
        self.set("y", f"{float(y):g}")
        return self

    @property
    def bounding_box(self) -> BoundingBox:
        """Returns the bounding box of the page."""
        return BoundingBox(
            (self.x, self.x + self.width), (self.y, self.y + self.height)
        )



"""
VNColor, VNGradient, pathStrokeStyle, basicStrokeStyle, styledElementData
"""

from __future__ import annotations

import colorsys
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

import inkex


class VNColor:
    """
    Represents a color with hex and alpha values.

    for styledTexts-fillColor-value,
    pathStrokeStyles-color and fills-color.
    """

    def __init__(self, color_dict: Dict):
        """
        Initializes the Color object from a dictionary containing color data.
        Handles both HSBA and RGBA formats.
        """
        self.hex: str = "#000000"
        self.alpha: float = 1.0

        color_data = self._extract_color_data(color_dict)
        if color_data is None:
            raise ValueError("Invalid color format")

        self.hex, self.alpha = color_data

    def _extract_color_data(self, color_dict: Dict) -> Optional[Tuple[str, float]]:
        """
        Extracts hex and alpha values from the given dictionary.
        Returns None if no valid color data is found.
        """
        rgba = color_dict.get("rgba")
        hsba = color_dict.get("hsba")

        if rgba:
            r, g, b, a = self._rgba_to_tuple(rgba)
        elif hsba:
            r, g, b, a = self._hsba_to_rgba_tuple(hsba)
        # Check for legacy HSB format
        elif "h" in color_dict and "s" in color_dict and "b" in color_dict:
            r, g, b, a = self._legacy_hsba_to_rgba_tuple(color_dict)
        else:
            return None

        hex_color = self._rgba_to_hex((r, g, b, a))
        return hex_color, a

    def _legacy_hsba_to_rgba_tuple(
        self, hsba: Dict
    ) -> Tuple[float, float, float, float]:
        """Converts an HSBA color to RGBA format."""
        hue = hsba.get("h", 0)
        saturation = hsba.get("s", 0)
        brightness = hsba.get("b", 0)
        alpha = hsba.get("a", 1)

        r, g, b = colorsys.hsv_to_rgb(hue, saturation, brightness)
        return r, g, b, alpha

    def _hsba_to_rgba_tuple(self, hsba: Dict) -> Tuple[float, float, float, float]:
        """Converts an HSBA color to RGBA format."""
        hue = hsba.get("hue", 0)
        saturation = hsba.get("saturation", 0)
        brightness = hsba.get("brightness", 0)
        alpha = hsba.get("alpha", 1)

        r, g, b = colorsys.hsv_to_rgb(hue, saturation, brightness)
        return r, g, b, alpha

    def _rgba_to_tuple(self, rgba: Dict) -> Tuple[float, float, float, float]:
        """Converts an RGBA color to tuple format."""
        red = rgba.get("red", 0)
        green = rgba.get("green", 0)
        blue = rgba.get("blue", 0)
        alpha = rgba.get("alpha", 1)

        return red, green, blue, alpha

    def _rgba_to_hex(self, rgba: Tuple[float, float, float, float]) -> str:
        """Converts RGBA tuple into str hex (#RRGGBB)."""
        r, g, b, _ = rgba
        r = int(r * 255)
        g = int(g * 255)
        b = int(b * 255)

        return f"#{r:02X}{g:02X}{b:02X}"


class VNGradient:
    """
    Represents a gradient with gradient and optional transform.
    """

    def __init__(
        self,
        fill_transform: Dict[str, Any],
        transform_matrix: Optional[List[float]],
        stops: List[Dict],
        typeRawValue: int,
    ):
        """
        Initializes the Gradient object from a Linearity Curve data.
        """
        self.gradient: inkex.Gradient = self._convert_gradient(
            tr=fill_transform, stops=stops, type_value=typeRawValue
        )
        self.transform: Optional[inkex.transforms.Transform] = None
        tr = inkex.transforms.Transform()
        if transform_matrix:
            tr.add_matrix(transform_matrix)
            self.transform = tr

    @staticmethod
    def _convert_gradient(
        tr: Dict[str, Any], stops: List[Dict], type_value: int
    ) -> inkex.Gradient:
        if type_value == 0:  # Linear Gradient
            gradient = inkex.LinearGradient()
            gradient.set("x1", tr["start"][0])
            gradient.set("y1", tr["start"][1])
            gradient.set("x2", tr["end"][0])
            gradient.set("y2", tr["end"][1])

        elif type_value == 1:  # Radial Gradient
            cx, cy = tr["start"]
            fx, fy = tr["end"]

            # Calculate radius (r) from start and end points. ??
            r = ((fx - cx) ** 2 + (fy - cy) ** 2) ** 0.5

            gradient = inkex.RadialGradient()
            gradient.set("cx", cx)
            gradient.set("cy", cy)
            gradient.set("r", r)

        gradient.set("gradientUnits", "userSpaceOnUse")

        # Add color stops
        for stop in stops:
            color = VNColor(color_dict=stop["color"])
            ratio = stop["ratio"]

            svg_stop = inkex.Stop()
            svg_stop.set("offset", ratio)
            svg_stop.style = inkex.Style(
                {
                    "stop-color": color.hex,
                    "stop-opacity": color.alpha,
                }
            )
            gradient.append(svg_stop)

        return gradient


@dataclass
class pathStrokeStyle:
    """
    Linearity Curve stroke format for path and text.
    """

    basicStrokeStyle: Optional[basicStrokeStyle]
    color: VNColor
    width: float


class basicStrokeStyle:
    """cap, dash, join, position"""

    def __init__(
        self, cap: int, dashPattern: Optional[List[float]], join: int, position: int
    ):
        self.cap: str = self._cap_to_svg(cap)
        self.dashPattern: List[float] = self._process_dash_pattern(dashPattern)
        self.join: str = self._join_to_svg(join)
        self.position: int = position

    @staticmethod
    def _process_dash_pattern(dashPattern: Optional[List[float]]) -> List[float]:
        if dashPattern is None:
            return [0]

        # delete trailing zeros
        while len(dashPattern) > 1 and dashPattern[-1] == 0:
            dashPattern.pop()

        return dashPattern

    @staticmethod
    def _cap_to_svg(cap):
        """Returns value for stroke-linecap attribute."""
        cap_map = {0: "butt", 1: "round", 2: "square"}
        return cap_map.get(cap, "butt")

    @staticmethod
    def _join_to_svg(join: int) -> str:
        """Returns value for stroke-linejoin attribute."""
        join_map = {
            0: "miter",
            1: "round",
            2: "bevel",
        }
        return join_map.get(join, "miter")


@dataclass
class styledElementData:
    """
    Stores style attributes for text for passing to
    read_vn_abst_path / read_vn_abst_text.
    """

    styled_data: Dict
    mask: int
    stroke: pathStrokeStyle
    color: Optional[VNColor]
    grad: Optional[VNGradient]



# coding=utf-8
#
# Copyright (C) 2008 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Base class for HGPL Encoding
"""

import re
import math

import inkex
from inkex.transforms import Transform, DirectedLineSegment, Vector2d
from inkex.bezier import cspsubdiv


class NoPathError(ValueError):
    """Raise that paths not selected"""


# Find the pen number in the layer number
FIND_PEN = re.compile(r"\s*pen\s*(\d+)\s*", re.IGNORECASE)
# Find the pen speed in the layer number
FIND_SPEED = re.compile(r"\s*speed\s*(\d+)\s*", re.IGNORECASE)
# Find pen force in the layer name
FIND_FORCE = re.compile(r"\s*force\s*(\d+)\s*", re.IGNORECASE)


class hpglEncoder(object):
    """HPGL Encoder, used by others"""

    def __init__(self, effect):
        """options:
        "resolutionX":float
        "resolutionY":float
        "pen":int
        "force:int
        "speed:int
        "orientation":string // "0", "90", "-90", "180"
        "mirrorX":bool
        "mirrorY":bool
        "center":bool
        "flat":float
        "overcut":float
        "toolOffset":float
        "precut":bool
        "autoAlign":bool
        """
        self.options = effect.options
        self.doc = effect.svg
        self.docWidth = effect.svg.viewbox_width
        self.docHeight = effect.svg.viewbox_height
        self.hpgl = ""
        self.divergenceX = "False"
        self.divergenceY = "False"
        self.sizeX = "False"
        self.sizeY = "False"
        self.dryRun = True
        self.lastPoint = [0, 0, 0]
        self.lastPen = -1
        self.lastSpeed = -1
        self.lastForce = -1
        self.offsetX = 0
        self.offsetY = 0
        # dots per inch to dots per user unit:

        self.scaleX = self.options.resolutionX / effect.svg.viewport_to_unit("1.0in")
        self.scaleY = self.options.resolutionY / effect.svg.viewport_to_unit("1.0in")
        scaleXY = (self.scaleX + self.scaleY) / 2

        # mm to dots (plotter coordinate system):
        self.overcut = (
            effect.svg.viewport_to_unit(str(self.options.overcut) + "mm") * scaleXY
        )
        self.toolOffset = (
            effect.svg.viewport_to_unit(str(self.options.toolOffset) + "mm") * scaleXY
        )

        # scale flatness to resolution:
        self.flat = self.options.flat / (
            1016 / ((self.options.resolutionX + self.options.resolutionY) / 2)
        )
        if self.toolOffset > 0.0:
            self.toolOffsetFlat = (
                self.flat / self.toolOffset * 4.5
            )  # scale flatness to offset
        else:
            self.toolOffsetFlat = 0.0
        self.mirrorX = -1.0 if self.options.mirrorX else 1.0
        self.mirrorY = 1.0 if self.options.mirrorY else -1.0
        # process viewBox attribute to correct page scaling
        self.viewBoxTransformX = 1
        self.viewBoxTransformY = 1
        viewBox = effect.svg.get_viewbox()
        if viewBox and viewBox[2] and viewBox[3]:
            self.viewBoxTransformX = self.docWidth / effect.svg.viewport_to_unit(
                effect.svg.add_unit(viewBox[2])
            )
            self.viewBoxTransformY = self.docHeight / effect.svg.viewport_to_unit(
                effect.svg.add_unit(viewBox[3])
            )

    def getHpgl(self):
        """Return the HPGL instructions"""
        # dryRun to find edges
        transform = Transform(
            [
                [self.mirrorX * self.scaleX * self.viewBoxTransformX, 0.0, 0.0],
                [0.0, self.mirrorY * self.scaleY * self.viewBoxTransformY, 0.0],
            ]
        )
        transform.add_rotate(int(self.options.orientation))

        self.vData = [
            ["", "False", 0],
            ["", "False", 0],
            ["", "False", 0],
            ["", "False", 0],
        ]
        self.process_group(self.doc, transform)
        if (
            self.divergenceX == "False"
            or self.divergenceY == "False"
            or self.sizeX == "False"
            or self.sizeY == "False"
        ):
            raise NoPathError("No paths found")
        # live run
        self.dryRun = False
        # move drawing according to various modifiers
        if self.options.autoAlign:
            if self.options.center:
                self.offsetX -= (self.sizeX - self.divergenceX) / 2
                self.offsetY -= (self.sizeY - self.divergenceY) / 2
        else:
            self.divergenceX = 0.0
            self.divergenceY = 0.0
            if self.options.center:
                if self.options.orientation == "0":
                    self.offsetX -= (self.docWidth * self.scaleX) / 2
                    self.offsetY += (self.docHeight * self.scaleY) / 2
                if self.options.orientation == "90":
                    self.offsetY += (self.docWidth * self.scaleX) / 2
                    self.offsetX += (self.docHeight * self.scaleY) / 2
                if self.options.orientation == "180":
                    self.offsetX += (self.docWidth * self.scaleX) / 2
                    self.offsetY -= (self.docHeight * self.scaleY) / 2
                if self.options.orientation == "270":
                    self.offsetY -= (self.docWidth * self.scaleX) / 2
                    self.offsetX -= (self.docHeight * self.scaleY) / 2
            else:
                if self.options.orientation == "0":
                    self.offsetY += self.docHeight * self.scaleY
                if self.options.orientation == "90":
                    self.offsetY += self.docWidth * self.scaleX
                    self.offsetX += self.docHeight * self.scaleY
                if self.options.orientation == "180":
                    self.offsetX += self.docWidth * self.scaleX
        if not self.options.center and self.toolOffset > 0.0:
            self.offsetX += self.toolOffset
            self.offsetY += self.toolOffset

        # initialize transformation matrix and cache
        transform = Transform(
            [
                [
                    self.mirrorX * self.scaleX * self.viewBoxTransformX,
                    0.0,
                    -float(self.divergenceX) + self.offsetX,
                ],
                [
                    0.0,
                    self.mirrorY * self.scaleY * self.viewBoxTransformY,
                    -float(self.divergenceY) + self.offsetY,
                ],
            ]
        )
        transform.add_rotate(int(self.options.orientation))
        self.vData = [
            ["", "False", 0],
            ["", "False", 0],
            ["", "False", 0],
            ["", "False", 0],
        ]
        # add move to zero point and precut
        if self.toolOffset > 0.0 and self.options.precut:
            if self.options.center:
                # position precut outside of drawing plus one time the tooloffset
                if self.offsetX >= 0.0:
                    precutX = self.offsetX + self.toolOffset
                else:
                    precutX = self.offsetX - self.toolOffset
                if self.offsetY >= 0.0:
                    precutY = self.offsetY + self.toolOffset
                else:
                    precutY = self.offsetY - self.toolOffset
                self.processOffset(
                    "PU",
                    Vector2d(precutX, precutY),
                    self.options.pen,
                    self.options.speed,
                    self.options.force,
                )
                self.processOffset(
                    "PD",
                    Vector2d(precutX, precutY + self.toolOffset * 8),
                    self.options.pen,
                    self.options.speed,
                    self.options.force,
                )
            else:
                self.processOffset(
                    "PU",
                    Vector2d(0, 0),
                    self.options.pen,
                    self.options.speed,
                    self.options.force,
                )
                self.processOffset(
                    "PD",
                    Vector2d(0, self.toolOffset * 8),
                    self.options.pen,
                    self.options.speed,
                    self.options.force,
                )
        # start conversion
        self.process_group(self.doc, transform)
        # shift an empty node in in order to process last node in cache
        if self.toolOffset > 0.0 and not self.dryRun:
            self.processOffset("PU", Vector2d(0, 0), 0, 0, 0)
        return self.hpgl

    def process_group(self, group, transform):
        """flatten layers and groups to avoid recursion"""
        for child in group:
            if not isinstance(child, inkex.ShapeElement):
                continue
            if child.is_visible():
                if isinstance(child, inkex.Group):
                    self.process_group(child, transform)
                elif isinstance(child, inkex.PathElement):
                    self.process_path(child, transform)
                else:
                    # This only works for shape elements (not text yet!)
                    new_elem = child.replace_with(child.to_path_element())
                    # Element is given composed transform b/c it's not added back to doc
                    new_elem.transform = child.composed_transform()
                    self.process_path(new_elem, transform)

    def get_pen_number(self, node):
        """Get pen number for node label (usually group)"""
        for parent in [node] + list(node.ancestors()):
            match = FIND_PEN.search(parent.label or "")
            if match:
                return int(match.group(1))
        return int(self.options.pen)

    def get_pen_speed(self, node):
        """Get pen speed for node label (usually group)"""
        for parent in [node] + list(node.ancestors()):
            match = FIND_SPEED.search(parent.label or "")
            if match:
                return int(match.group(1))
        return int(self.options.speed)

    def get_pen_force(self, node):
        """Get pen force for node label (usually group)"""
        for parent in [node] + list(node.ancestors()):
            match = FIND_FORCE.search(parent.label or "")
            if match:
                return int(match.group(1))
        return int(self.options.force)

    def process_path(self, node, transform):
        """Process the given element into a plotter path"""
        pen = self.get_pen_number(node)
        speed = self.get_pen_speed(node)
        force = self.get_pen_force(node)

        path = (
            node.path.to_absolute()
            .transform(node.composed_transform())
            .transform(transform)
            .to_superpath()
        )
        if path:
            cspsubdiv(path, self.flat)
            # path to HPGL commands
            oldPosX = 0.0
            oldPosY = 0.0
            for singlePath in path:
                non_empty_subpath = False
                for i, singlePathPoint in enumerate(singlePath):
                    posX, posY = singlePathPoint[1]
                    # check if point is repeating, if so, ignore
                    if int(round(posX)) != int(round(oldPosX)) or int(
                        round(posY)
                    ) != int(round(oldPosY)):
                        non_empty_subpath = True
                        self.processOffset(
                            "PU" if not i else "PD",
                            Vector2d(posX, posY),
                            pen,
                            speed,
                            force,
                        )
                        oldPosX = posX
                        oldPosY = posY
                # perform overcut
                if self.overcut > 0.0 and not self.dryRun and non_empty_subpath:
                    # check if last and first points are the same, otherwise the path
                    # is not closed and no overcut can be performed
                    if int(round(oldPosX)) == int(round(singlePath[0][1][0])) and int(
                        round(oldPosY)
                    ) == int(round(singlePath[0][1][1])):
                        overcutLength = 0
                        for singlePathPoint in singlePath:
                            posX, posY = singlePathPoint[1]
                            # check if point is repeating, if so, ignore
                            if int(round(posX)) != int(round(oldPosX)) or int(
                                round(posY)
                            ) != int(round(oldPosY)):
                                overcutLength += (
                                    Vector2d(posX, posY) - (oldPosX, oldPosY)
                                ).length
                                if overcutLength >= self.overcut:
                                    newEndPoint = self.changeLength(
                                        Vector2d(oldPosX, oldPosY),
                                        Vector2d(posX, posY),
                                        -(overcutLength - self.overcut),
                                    )
                                    self.processOffset(
                                        "PD", newEndPoint, pen, speed, force
                                    )
                                    break
                                self.processOffset(
                                    "PD", Vector2d(posX, posY), pen, speed, force
                                )
                                oldPosX = posX
                                oldPosY = posY

    def changeLength(self, p1, p2, offset):
        """change length of line"""
        if p1.x == p2.x and p1.y == p2.y:  # abort if points are the same
            return p1
        return Vector2d(DirectedLineSegment(p2, p1).point_at_length(-offset))

    def processOffset(self, cmd, point, pen, speed, force):
        """Calculate offset correction"""
        if self.toolOffset == 0.0 or self.dryRun:
            self.storePoint(cmd, point, pen, speed, force)
        else:
            # insert data into cache
            self.vData.pop(0)
            self.vData.insert(3, [cmd, point, pen, speed, force])
            # decide if enough data is available
            if self.vData[2][1] != "False":
                if self.vData[1][1] == "False":
                    self.storePoint(
                        self.vData[2][0],
                        self.vData[2][1],
                        self.vData[2][2],
                        self.vData[2][3],
                        self.vData[2][4],
                    )
                else:
                    # perform tool offset correction (It's a *tad* complicated, if you want
                    #                     to understand it draw the data as lines on paper)
                    if self.vData[2][0] == "PD":
                        # If the 3rd entry in the cache is a pen down command,
                        #             make the line longer by the tool offset
                        pointThree = self.changeLength(
                            self.vData[1][1], self.vData[2][1], self.toolOffset
                        )
                        self.storePoint(
                            "PD",
                            pointThree,
                            self.vData[2][2],
                            self.vData[2][3],
                            self.vData[2][4],
                        )
                    elif self.vData[0][1] != "False":
                        # Elif the 1st entry in the cache is filled with data and the 3rd entry
                        #   is a pen up command shift the 3rd entry by the current tool offset
                        #   position according to the 2nd command
                        pointThree = self.changeLength(
                            self.vData[0][1], self.vData[1][1], self.toolOffset
                        )
                        pointThree = self.vData[2][1] - (self.vData[1][1] - pointThree)
                        self.storePoint(
                            "PU",
                            pointThree,
                            self.vData[2][2],
                            self.vData[2][3],
                            self.vData[2][4],
                        )
                    else:
                        # Else just write the 3rd entry
                        pointThree = self.vData[2][1]
                        self.storePoint(
                            "PU",
                            pointThree,
                            self.vData[2][2],
                            self.vData[2][3],
                            self.vData[2][4],
                        )
                    if self.vData[3][0] == "PD":
                        # If the 4th entry in the cache is a pen down command guide tool to next
                        #           line with a circle between the prolonged 3rd and 4th entry
                        originalSegment = DirectedLineSegment(
                            self.vData[2][1], self.vData[3][1]
                        )
                        if originalSegment.length >= self.toolOffset:
                            pointFour = self.changeLength(
                                originalSegment.end,
                                originalSegment.start,
                                -self.toolOffset,
                            )
                        else:
                            pointFour = self.changeLength(
                                originalSegment.start,
                                originalSegment.end,
                                self.toolOffset - originalSegment.length,
                            )
                        # get angle start and angle vector
                        angleStart = DirectedLineSegment(
                            self.vData[2][1], pointThree
                        ).angle
                        angleVector = (
                            DirectedLineSegment(self.vData[2][1], pointFour).angle
                            - angleStart
                        )
                        # switch direction when arc is bigger than 180°
                        if angleVector > math.pi:
                            angleVector -= math.pi * 2
                        elif angleVector < -math.pi:
                            angleVector += math.pi * 2
                        # draw arc
                        if angleVector >= 0:
                            angle = angleStart + self.toolOffsetFlat
                            while angle < angleStart + angleVector:
                                self.storePoint(
                                    "PD",
                                    self.vData[2][1]
                                    + self.toolOffset
                                    * Vector2d(math.cos(angle), math.sin(angle)),
                                    self.vData[2][2],
                                    self.vData[2][3],
                                    self.vData[2][4],
                                )
                                angle += self.toolOffsetFlat
                        else:
                            angle = angleStart - self.toolOffsetFlat
                            while angle > angleStart + angleVector:
                                self.storePoint(
                                    "PD",
                                    self.vData[2][1]
                                    + self.toolOffset
                                    * Vector2d(math.cos(angle), math.sin(angle)),
                                    self.vData[2][2],
                                    self.vData[2][3],
                                    self.vData[2][4],
                                )
                                angle -= self.toolOffsetFlat
                        self.storePoint(
                            "PD",
                            pointFour,
                            self.vData[3][2],
                            self.vData[2][3],
                            self.vData[2][4],
                        )

    def storePoint(self, command, point, pen, speed, force):
        x = int(round(point.x))
        y = int(round(point.y))
        # skip when no change in movement
        if (
            self.lastPoint[0] == command
            and self.lastPoint[1] == x
            and self.lastPoint[2] == y
        ):
            return
        if self.dryRun:
            # find edges
            if self.divergenceX == "False" or x < self.divergenceX:
                self.divergenceX = x
            if self.divergenceY == "False" or y < self.divergenceY:
                self.divergenceY = y
            if self.sizeX == "False" or x > self.sizeX:
                self.sizeX = x
            if self.sizeY == "False" or y > self.sizeY:
                self.sizeY = y
        else:
            # store point
            if not self.options.center:
                # only positive values are allowed (usually)
                if x < 0:
                    x = 0
                if y < 0:
                    y = 0
            # select correct pen
            if self.lastPen != pen:
                self.hpgl += ";PU;SP%d" % pen
            if self.lastSpeed != speed:
                if speed > 0:
                    self.hpgl += ";VS%d" % speed
            if self.lastForce != force:
                if force > 0:
                    self.hpgl += ";FS%d" % force
            # do not repeat command
            if command == "PD" and self.lastPoint[0] == "PD" and self.lastPen == pen:
                self.hpgl += ",%d,%d" % (x, y)
            else:
                self.hpgl += ";%s%d,%d" % (command, x, y)
            self.lastPen = pen
            self.lastSpeed = speed
            self.lastForce = force
        self.lastPoint = [command, x, y]
